from skgaze.common_scanpath import STA
from skgaze.common_scanpath import eMine
from skgaze.common_scanpath import Dotplots
from skgaze.core.Aoi import Aoi
from skgaze.core.Participant import Participant
import csv
import math

class Dataset:

    def __init__(self,dataset_file_path,aoi_file_path=False,filtered_dataset=False):
        self.dataset_file_path = dataset_file_path
        self.aoi_file_path = aoi_file_path
        self.aoi_array = self.load_aoi()
        self.participants = []
        self.last_participant_name = ""
        self.filtered_dataset = filtered_dataset
        self.startTimestamp = None
        self.lastTimestamp = None
       
        self.data = self.load_dataset()


    def load_dataset(self):
        data = open(self.dataset_file_path, 'r')
        reader = csv.DictReader(data, delimiter='\t')
        current_fix = 0
        act_file_data = []
        all_file_data = []
        for row in reader:
                    # if self.website_name not in row["MediaName"].split(" ")[0] :  # ignore non-recording data point
                    #     continue
          
                
            if not self.filtered_dataset and (not row["ValidityLeft"] or not row["ValidityRight"] or not row["FixationPointX (MCSpx)"] or not \
            row["FixationPointY (MCSpx)"]):  # ignore data point with no information
                continue
            if row["GazeEventType"] != "Fixation" or current_fix == int(row["FixationIndex"].split(".")[0]):
                            # if not a fixation or the current fixation
                continue
                        # clear data on visit of first line of new participant
            if self.last_participant_name !=row["ParticipantName"] :
                if len(act_file_data) != 0:
                    participant_identifier = self.last_participant_name
                    self.participants.append(Participant(participant_identifier,act_file_data,self.aoi_array,self.startTimestamp,self.lastTimestamp))
                act_file_data = []
                self.startTimestamp = row["RecordingTimestamp"]
                self.last_participant_name = row["ParticipantName"]
            self.lastTimestamp = row["RecordingTimestamp"]
            data = [str(int(row["FixationIndex"].split(".")[0]) - 1),
                                        row["GazeEventDuration"],
                                        row["FixationPointX (MCSpx)"],
                                        row["FixationPointY (MCSpx)"],
                                        row["MediaName"].split(" ")[0],
                                        row["RecordingTimestamp"],
                                        row["LocalTimeStamp"],
                                        "0.0."+row["LocalTimeStamp"].split(".")[1],
                                        row["FixationIndex"],
                                        row["GazePointLeftX (ADCSpx)"],
                                        row["GazePointLeftY (ADCSpx)"],
                                        row["CamLeftX"],
                                        row["CamLeftY"],
                                        row["DistanceLeft"],
                                        row["PupilLeft"],
                                        row["ValidityLeft"],
                                        row["GazePointRightX (ADCSpx)"],
                                        row["GazePointRightY (ADCSpx)"],
                                        row["CamRightX"],
                                        row["CamRightY"],
                                        row["DistanceRight"],
                                        row["PupilRight"],
                                        row["ValidityRight"],
                                        row["FixationIndex"],
                                        row["GazePointX (ADCSpx)"],
                                        row["GazePointY (ADCSpx)"],
                                        "ScreenRecStarted",
                                        8192,
                                        0,
                                        0,
                                        "",
                                        "ScreenRec",
                                        0,
                                        row["MediaWidth"],
                                        row["MediaHeight"],
                                        row["MediaPosX (ADCSpx)"],row["MediaPosY (ADCSpx)"],row["FixationPointX (MCSpx)"],
                                        row["FixationPointY (MCSpx)"], row["GazeEventDuration"],0,
                                        "Content",
                                        "",row["GazePointX (MCSpx)"],
                                        row["GazePointY (MCSpx)"],row["RecordingTimestamp"],
                                        row["EyeTrackerTimestamp"]
                                     ]
            act_file_data.append(data)
            all_file_data.append(data)
            
            
        
        self.participants.append(Participant(self.last_participant_name,act_file_data,self.aoi_array,self.startTimestamp,self.lastTimestamp))
        return all_file_data

    def load_aoi(self):
        if(not self.aoi_file_path):
            return
        aoi_data = open(self.aoi_file_path)
        aoi_array = []
        for line in aoi_data:
            aoi = line.split()
            aoi_array.append(Aoi(aoi[0],aoi[1],aoi[2],aoi[3],aoi[4],aoi[5]))
        return aoi_array

    def create_raw_sequences(self):
        raw_sequences = {}
        for participant in self.participants:
            raw_sequences[participant.id] = participant.scanpath
        return raw_sequences

    def info(self):
        print("Number of participants: ",len(self.participants))
        print("Number of Areas of Interest (AIO): ", len(self.aoi_array))
        print("Filtered dataset: ", self.filtered_dataset)
    
    # def create_aoi_array_simplified(self):
    #     for aoi in self.create_aoi_array()




class Participant:

    def __init__(self,id,fixations,aoi_array,startTimestamp, lastTimestamp):
        self.id = id
        self.fixations = fixations
        self.scanpath = []
        self.aoi_array = aoi_array
        self.startTimestamp = startTimestamp
        self.lastTimestamp = lastTimestamp
        self.load_scanpaths()


    def load_scanpaths(self):
        
        for row in self.fixations:
            for aoi in self.aoi_array:
                aoi_x = float(aoi.x)
                aoi_width = float(aoi.width)
                aoi_y = float(aoi.y)
                aoi_height = float(aoi.height)
                fixation_x = float(row[2])
                fixation_y = float(row[3])
                if ((aoi_x  <= fixation_x < aoi_x + aoi_width) and \
                        (aoi_y   <= fixation_y < aoi_y  + aoi_height)):
                    self.scanpath.append([aoi.aoi_char,row[1]])



import copy
	
from skgaze.similarity_algorithms import similarity
from skgaze.similarity_algorithms import longest_common_subsequence
	
class eMine:

    def __init__(self,dataset):
        self.raw_sequences = dataset.create_raw_sequences()

    def simplify_scanpath(self,scanpath):
	    res_str = ''
	
	    for i in range(0, len(scanpath)):
	        if i == 0:
	            res_str += scanpath[i]
	        elif i > 0 and scanpath[i] != scanpath[i - 1]:
	            res_str += scanpath[i]
	
	    return res_str
    def format_sequences(self,raw_sequences):
	    """
	    {'01': [[A, 150], [B, 250]], '02': ...} gets transformed into:
	    [{'identifier': '01', 'fixations': [[A, 150], [B, 250]]}, {'identifier': '02' ... }]
	    """
	    formatted_sequences = []
	    keys = list(raw_sequences.keys())
	    for it in range(0, len(raw_sequences)):
	        act_rec = {
	            'identifier': keys[it],
	            'fixations': raw_sequences[keys[it]]
	        }
	        formatted_sequences.append(act_rec)
	
	    return formatted_sequences
	
	# eMINE algorithm (https://bop.unibe.ch/index.php/JEMR/article/view/2430)
    def run_emine(self):
	    """
	    Args:
	        raw_sequences: a Python dict of lists - {'ID1': [['F', '383'], ['G', '150']], 'ID2': .. }
	    Returns:
	        identifier: for client-side purposes
	        fixations: a list of lists representing the common scanpath - [['A', 150], ['B', 500] .. ]
	        similarity: a dict containing similarity of individual scanpaths to the common one - {'ID1': 66.66, 'ID2': ... }
	    """
	
	    formatted_sequences = self.format_sequences(self.raw_sequences)
	
	    # Store scanpaths as an array of string-converted original scanpaths (for calculating LCS etc.)
	    scanpath_strs = similarity.convert_to_str_array(formatted_sequences)
	    scanpath_strs_set = copy.deepcopy(scanpath_strs)
	
	    # For determining get_edit_distance distance we need a pure string version of the common scanpath ('ABC')
	    common_scanpath_str = ''
	
	    # Process until there is only 1 (common) scanpath left in the set
	    while len(scanpath_strs_set):
	        # Calculate the mutual similarities if there are at least 2 scanpaths in the set
	        if len(scanpath_strs_set) > 1:
	            similarity.calc_mutual_similarity(scanpath_strs_set)
	        else:
	            common_scanpath_str = self.simplify_scanpath(scanpath_strs_set[0]['raw_str'])
	            break
	
	        # Get the two most similar scanpaths
	        most_similar_pair = similarity.get_most_similar_pair(scanpath_strs_set)
	
	        lcs = longest_common_subsequence.get_longest_common_subsequence(
	            most_similar_pair[0]['raw_str'],
	            most_similar_pair[1]['raw_str'],
	        )
	
	        # Common scanpath does not exist
	        if not lcs:
	            break
	
	        # Remove the most similar pair of scanpaths from the set
	        similarity.rem_scanpath_strs_by_id(
	            scanpath_strs_set,
	            [most_similar_pair[0]['identifier'], most_similar_pair[1]['identifier']]
	        )
	        # Insert their longest common substring instead
	        scanpath_strs_set.append({
	            'raw_str': lcs,
	            'identifier': most_similar_pair[0]['identifier']
	        })
	
	    common_scanpath_arr = []
	    for i in range(0, len(common_scanpath_str)):
	        common_scanpath_arr.append([common_scanpath_str[i], 0])
	
	    res_data = {
	        'identifier': 'eMINE',
	        'fixations': common_scanpath_arr,
	        'similarity': similarity.calc_similarity_to_common(scanpath_strs, common_scanpath_str)
	    }
	
	    return res_data
	


"""
UBC Eye Movement Data Analysis Toolkit

Author: Nicholas FitzGerald - nicholas.fitzgerald@gmail.com
Modified by: Samad Kardan
             Oliver Schmid - oliver.schmd@gmail.com

Basic data structures used in EMDAT
"""
from warnings import warn

class Datapoint():
    """
    A class that holds the information for one eye gaze data sample (one line of data logs) 
    
    Attributes:
        segid: a string indicating the Segment that this Datapoint belongs to
        is_valid: a boolean indicating whether this sample is valid
    
        Please refer to the Tobii manual for the description of the rest of the attributes
    """

    def __init__(self, tobii_line, data=None):
        """
        Initializes a Datapoint from either a line of gaze data from "all-Data.tsv"
        or the equivalent data in array form.

        Args:
            tobii_line: a line of gaze data from "all-Data.tsv"
            data: An already parsed line of Tobii data

        Yields:
            a Datapoint object
        """
        if data == None:
            data = tobii_line[5:]
        if data:
            [self.timestamp, self.datetimestamp, self.datetimestampstartoffset, self.number, self.gazepointxleft, self.gazepointyleft, self.camxleft, self.camyleft, 
             self.distanceleft, self.pupilleft, self.validityleft, self.gazepointxright, self.gazepointyright, self.camxright, self.camyright, self.distanceright, 
             self.pupilright, self.validityright, self.fixationindex, self.gazepointx, self.gazepointy, self.event, self.eventkey, self.data1, self.data2, self.descriptor, 
             self.stimuliname, self.stimuliid, self.mediawidth, self.mediaheight, self.mediaposx, self.mediaposy, self.mappedfixationpointx, self.mappedfixationpointy, 
             self.fixationduration, self.aoiids, self.aoinames, self.webgroupimage, self.mappedgazedatapointx, self.mappedgazedatapointy, self.microsecondtimestamp, 
             self.absolutemicrosecondtimestamp] = data
            if self.number != None:
                self.is_None = False
                self.segid = None
                self.is_valid = (int(self.validityright) < 2 or int(self.validityleft) < 2)
            else:
                self.is_None = True
        else:
            self.is_None = True

    def set_segid(self,segid):
        """Sets the "Segment" id for this Datapoint
        
        Args:
            segid: a string containing the "Segment" id
        """
        self.segid = segid

    def get_segid(self):
        """Returns the "Segment" id for this Datapoint
            
        Returns:
            a string containing the "Segment" id
            
        Raises:
            Exception: if the segid is not set before reading it an Exception will be thrown
        """
        if self.segid != None:
            return self.segid
        raise Exception('The segid is accessed before setting the initial value in a datapoint!')


class Fixation():
    """
    A class that holds the information for one Fixation representing one line in a "Fixation-Data.tsv" file
    
    Attributes:
        segid: a string indicating the Segment that this Datapoint belongs to
    
        Please refer to the Tobii manual for the description of the rest of the attributes
    """

    def __init__(self, fix_point, media_offset = (0, 0)):
        """Inits Fixation class with a line of gaze data from a "Fixation-Data.tsv" file
        
        Args:
            fix_point: a string containing one line read from a "Fixation-Data.tsv" file        
            media_offset: the coordinates of the top left corner of the window
                showing the interface under study. (0,0) if the interface was
                in full screen (default value)
            
        Yields:
            a Fixation object
        """
        data = [fix_point[0], fix_point[5], fix_point[1], fix_point[2],fix_point[3]]
        #fix_point = fix_point.replace('\t\r\n','')
        [self.fixationindex, self.timestamp, self.fixationduration, self.mappedfixationpointx, self.mappedfixationpointy] = data
        self.fixationindex = cast_int(self.fixationindex)
        self.timestamp = cast_int(self.timestamp)
        self.fixationduration = cast_int(self.fixationduration)
        if self.fixationduration == 0:
            warn("A zero duration Fixation!")
        (media_offset_x, media_offset_y) = media_offset
        self.mappedfixationpointx = cast_int(self.mappedfixationpointx) - media_offset_x
        self.mappedfixationpointy = cast_int(self.mappedfixationpointy) - media_offset_y
        self.segid = None
        
    def set_segid(self,segid):
        """Sets the "Segment" id for this Fixation
        
        Args:
            segid: a string containing the "Segment" id
        """
        self.segid = segid

    def get_segid(self):
        """Returns the "Segment" id for this Fixation
            
        Returns:
            a string containing the "Segment" id
            
        Raises:
            Exception: if the segid is not set before reading it an Exception will be thrown
        """
        if self.segid != None:
            return self.segid
        raise Exception('The segid is accessed before setting the initial value in a fixation point!')


class Event():
    """
    A class that holds the information for one Event representing one line in the an "Event-Data.tsv" file
    
    Please refer to the Tobii manual for the description of the attributes
    """
    def __init__(self, eventstr):
        """Inits Event class with a line of gaze data from an "Event-Data.tsv" file
        
        Args:
            eventstr: a string containing one line read from an "Event-Data.tsv" file
            
        Yields:
            an Event object
        """
        #Timestamp    Event    EventKey    Data1    Data2    Descriptor
        #print eventstr.split('\t')
        [self.timestamp, self.event, self.eventKey, self.data1, self.data2,self.descriptor,_] = eventstr.split('\t')
        self.timestamp = cast_int(self.timestamp)
        self.eventKey = cast_int(self.eventKey)


def cast_int(str):
    """a helper method for converting strings to their integer value
    
    Args:
        str: a string containing a number
    
    Returns:
        the integer value of the string given or None if not an integer
    """
    if str!="" and str!=None:
        v = int(str)
    else:
        v = None
    return v
#    try:
#        v = int(str)
#    except ValueError:
#        v = None
#    return v


def cast_float(str):
    """a helper method for converting strings to their float value
    
    Args:
        str: a string containing a number
    
    Returns:
        the float value of the string given or None if not a float
    """
    if str!="" and str!=None:
        v = float(str)
    else:
        v = None
    return v
#    try:
#        v = float(str)
#    except ValueError:
#        v = None
#    return v




class Aoi:
    def __init__(self,type_of_element,x,width,y,height,aoi_char):
        self.type_of_element = type_of_element
        self.x = x
        self.width = width
        self.y = y
        self.height = height
        self.aoi_char = aoi_char



"""
UBC Eye Movement Data Analysis Toolkit
The Generic Area of Interest Classes
Created on 2011-08-26

@author: skardan

In EMDAT, the bounderies of an Area of Interest (AOI) is defined as a polygon on the screen. You can 
optionally define a second polygone inside the first polygone to be excluded from an AOI.
An AOI can be always active (a global AOI) or can be active during certain time intervals.
In order to calculate the features for an AOI instance, you need to create an AOI_Stat instance and
map it to a target AOI object by passing it to the AOI_Stat constructor. The resulting AOI_Stat
will calculate all features related to the given AOI and store them for later reference 
"""
from skgaze.EMDAT.utils import *
import skgaze.EMDAT.params as params
from warnings import warn




class AOI():

    def __init__(self, aid, polyin, polyout=[], timeseq=[]):
        """Inits AOI class
        Args:
            aid: AOI id
            
            polyin: the polygon defining the boundaries of the AOI in form of a list of (x,y) tuples
            
            polyout: optional polygon inside the boundaries of the AOI that is not part of 
                the AOI in form of a list of (x,y) tuples
            
            timeseq: the time sequence of the format [(start1, end1), (start2, end2), ...] that 
                specifies the intervals when this AOI is active
            
        Yields:
            an AOI object
        """
        self.aid = aid
        self.polyin = polyin
        self.polyout = polyout
        self.timeseq = timeseq
#            self.partial = True
    def set_coordinates(self, polyin, polyout=[]):
        """Sets the coordiantes of the AOI
        
        Args:
            polyin: the polygon defining the bounderies of the AOI in form of a list of (x,y) tuples
            polyout: optional polygon inside the bounderies of the AOI that is not part of the AOI 
                in form of a list of (x,y) tuples
        """
        
        self.polyin = polyin
        self.polyout = polyout
            
    def is_active(self,start,end):
        """Determines if an AOI is active during the whole given time interval 

                
        Args:
            start: time interval start
            end: time interval end
            
        Returns:
            true if the AOI is always active within the given time interval 
        """
        if start == -1:
            return False
        if self.timeseq:
            for intr in self.timeseq:
                if (start>=intr[0] and start<intr[1])or(end>intr[0] and end<=intr[1]):
                    return True
                elif (start<intr[0] and start<intr[1])and(end>intr[0] and end>intr[1]):
                    warn("Incorrect definition of Dynamic AOI and Segments, AOI info not calculated for AOI:"+self.aid)
            return False #not active
        else:
            return True #global AOI
        
    def is_active_partition(self,start,end):
        """Determines if an AOI is partially active during a given time interval
        
        If the AOI is active at any sub-interval of the given time interval returns true
        if such sub-interval exists it also returns its start and end that AOI is active otherwise returns False, []
        
        Args:
            start: time interval start
            end: time interval start
            
        Returns:
            A boolean for whether the AOI is active or not 
            ovelap_part: The subset of the time interval [sub_start,sub_end] that AOI is active or 
            [] if it is active during the whole interval or not active at all. 
        """ 
        #if (end - start)== 0:
        if start == -1:
            return False, []
        if params.DEBUG:
            print("in:",self.aid)
        ovelap_part = []
        if self.timeseq:
            for intr in self.timeseq:
                if (start>=intr[0] and start<intr[1])or(end>intr[0] and end<=intr[1]):
                    if (start>=intr[0] and end<=intr[1]):
                        ovelap_part=[]
                    else:
                        if params.DEBUG:
                            print("partial:",start,end,":",intr[0],intr[1])
                        ovstart = max(start,intr[0])
                        ovend  = min(end,intr[1])
                        ovelap_part = [ovstart,ovend]
                    return True, ovelap_part
                elif (start<intr[0] and start<intr[1])and(end>intr[0] and end>intr[1]):
                    warn("Incorrect definition of Dynamic AOI and Segments, AOI info not calculated for AOI:"+self.aid)

            return False, [] #not active
        else:
            return True, [] #global AOI
    
                   
class AOI_Stat():
    """Methods of AOI_Stat calculate and store all features related to the given AOI object
    """

    def __init__(self,aoi,seg_fixation_data, starttime, endtime, active_aois):
        """Inits AOI_Stat class
        
        Args:
            aoi: the aoi object for which the statistics are calculated
            seg_fixation_data:
            starttime:
            endtime:
            active_aois:list of the AOI objects that will be used for calculating the transitions between this AOI and other AOIs 
            
        Yields:
            an AOI_Stat object
        """
        self.aoi = aoi
        self.isActive, partition = self.aoi.is_active_partition(starttime, endtime)
        self.aoi
        if not(self.isActive):
            return
        self.isActive = True
        
        if partition:
            if params.DEBUG:
                print("partition",partition)
            _,st,en = get_chunk(seg_fixation_data, 0, partition[0],partition[1])
            fixation_data = seg_fixation_data[st:en]
            if params.DEBUG:
                print("len(seg_fixation_data)",seg_fixation_data)
                print("len(fixation_data)",fixation_data)
        else:  #global AOI (alaways active)
            fixation_data = seg_fixation_data 
        fixation_indices = filter(lambda i: _fixation_inside_aoi(fixation_data[i],self.aoi.polyin, self.aoi.polyout), range(len(fixation_data)))

        fixations = list(map(lambda i: fixation_data[i], fixation_indices))
    
        self.features = {}

        numfixations = len(fixations)
        self.features['numfixations'] = numfixations
        self.features['longestfixation'] = -1
        self.features['timetofirstfixation'] = -1
        self.features['timetolastfixation'] = -1
        self.features['proportionnum'] = 0
        totaltimespent = sum(map(lambda x: x.fixationduration, fixations))
        self.features['totaltimespent'] = totaltimespent 
        
        self.features['proportiontime'] = float(totaltimespent)/(float(endtime) - float(starttime))
        if numfixations > 0:
            self.features['longestfixation'] = max(map(lambda x: x.fixationduration,
            fixations))
            self.features['timetofirstfixation'] = fixations[0].timestamp - float(starttime)
            self.features['timetolastfixation'] = fixations[-1].timestamp - float(starttime)
            self.features['proportionnum'] = float(numfixations)/len(fixation_data)
            self.features['fixationrate'] = numfixations / float(totaltimespent)
        else:
            self.features['longestfixation'] = 0
            self.features['timetofirstfixation'] = 0
            self.features['timetolastfixation'] = 0
            self.features['proportionnum'] = 0
            self.features['fixationrate'] = 0
                
                
            
        #calculating the transitions to and from this AOI and other active AOIs at the moment
        for aoi in active_aois:
            aid = aoi.aid
            self.features['numtransto_%s'%(aid)] = 0
            self.features['numtransfrom_%s'%(aid)] = 0

        sumtransto = 0
        sumtransfrom = 0
        for i in fixation_indices:
            if i > 0:
                for aoi in active_aois:
                    aid = aoi.aid
                    polyin = aoi.polyin
                    polyout = aoi.polyout
                    key = 'numtransfrom_%s'%(aid)
                    #self.features[key] = 0 #????? Samad
                    if _fixation_inside_aoi(fixation_data[i-1], polyin, polyout):
                        self.features[key] += 1
                        sumtransfrom += 1
            if i < len(fixation_data)-2:
                for aoi in active_aois:
                    aid = aoi.aid
                    polyin = aoi.polyin
                    polyout = aoi.polyout
                    key = 'numtransto_%s'%(aid)
                    #self.features[key] = 0 #????? Samad
                    if _fixation_inside_aoi(fixation_data[i+1], polyin, polyout):
                        self.features[key] += 1
                        sumtransto += 1

        for aoi in active_aois:
            aid = aoi.aid
            if sumtransto > 0:
                val = self.features['numtransto_%s'%(aid)]
                self.features['proptransto_%s'%(aid)] = float(val) / sumtransto
            else:
                self.features['proptransto_%s'%(aid)] = 0

            if sumtransfrom > 0:
                val = self.features['numtransfrom_%s'%(aid)]
                self.features['proptransfrom_%s'%(aid)] = float(val) / sumtransfrom
            else:
                self.features['proptransfrom_%s'%(aid)] = 0
        self.total_tans_to = sumtransto
        self.total_tans_from = sumtransfrom
        ###endof trnsition calculation


    def get_features(self, featurelist = None):
        """Returns the list of names and values of features for this AOI_Stat object
        
        Args:
            featurelist: optional list of features. If equal to None the full set of all features will be returned
            
        Returns:
            featnames: a list of feature names sorted alphabetically
            featvals: a corresponding list of feature values
            e.g.
            featnames = ['fixationrate', 'length', 'meanabspathangles']
            featvals  = [0.00268522882294', '1529851', '1.60354714212']
        
        """
        if featurelist == []:
            return [], []
        elif not featurelist:   #all features
            featnames = self.features.keys()
        else:                   #a list was given
            featnames = []
            for name in featurelist:
                if name == 'numtransto':
                    featnames += filter(lambda x: x[:10] == 'numtransto', self.features.keys())
                elif name == 'numtransfrom':
                    featnames += filter(lambda x: x[:12] == 'numtransfrom', self.features.keys())
                elif name == 'proptransfrom':
                    featnames += filter(lambda x: x[:11] == 'proptransto', self.features.keys())
                elif name == 'proptransto':
                    featnames += filter(lambda x: x[:13] == 'proptransfrom', self.features.keys())
                elif name in self.features.keys():
                    featnames.append(name)
                else:
                    raise Exception('AOI %s has no such feature: %s'%(self.aoi.aid,
                    name))

        featnames = sorted(featnames)

        featvals = map(lambda x: self.features[x], featnames)

        return featnames, featvals

    def print_(self):
        """Prints the list of features and their values for this AOI_Stat object
     
        """

        print("AOI ID:",self.aoi.aid)
        fn,fv = self.get_features()
        for i in range(len(fn)):
            print(fn[i],':',fv[i])
        print()
            

def _fixation_inside_aoi(fixation, polyin, polyout):
    """Helper function that checks if a fixation object is inside the AOI described by external polygon polyin and the internal polygon polyout.
    
    Fixation object is inside AOI if it is inside polyin but outside polyout
    
    Args:
        fixation: A Fixation object
        polyin: the external polygon in form of a list of (x,y) tuples
        polyout: the internal polygon in form of a list of (x,y) tuples
    
    Returns: 
        A boolean for whether the Fixation is inside the AOI or not
    """
    return point_inside_polygon(fixation.mappedfixationpointx,
    fixation.mappedfixationpointy, polyin) and not point_inside_polygon(fixation.mappedfixationpointx,
    fixation.mappedfixationpointy, polyout)     
         



def get_edit_distance(s1, s2):
    # Source: https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance

    if len(s1) < len(s2):
        return get_edit_distance(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            # j+1 instead of j since previous_row and current_row are one character longer than s2
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)

            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]



from skgaze.similarity_algorithms import levenshtein_distance

def calc_similarity(scanpath1, scanpath2, algorithm=False):
    """ Calculates similarity between two scanpath strings represented as percentage. """
    if not algorithm:
        algorithm = levenshtein_distance.get_edit_distance
    
    # Calculate the similarity value based on the edit distance of given scanpaths
    edit_distance = algorithm(scanpath1, scanpath2)
    similarity = 1 - (edit_distance / (len(scanpath1) if len(scanpath1) > len(scanpath2) else len(scanpath2)))

    # Return similarity as percentage
    return similarity * 100

def calc_mutual_similarity(scanpath_strs):
    """ Calculates mutual similarity between all scanpaths of the formatted array. """

    for i_first in range(0, len(scanpath_strs)):
        # Each scanpath has a similarity object - similarity[id] represents
        # the level of similarity to the scanpath identified by id
        scanpath1 = scanpath_strs[i_first]

        # If the similarity object of first scanpath does not exist yet - create it
        if not scanpath1.get('similarity'):
            scanpath1['similarity'] = {}

        for i_second in range(i_first + 1, len(scanpath_strs)):
            scanpath2 = scanpath_strs[i_second]

            # Get the IDs of the current scanpath pair
            identifier_first = scanpath1['identifier']
            identifier_second = scanpath2['identifier']

            # Performance boost: skip if the similarity between current scanpath pair has already been calculated
            if scanpath1.get('similarity') and scanpath1['similarity'].get(identifier_second) and \
                    scanpath2.get('similarity') and scanpath2['similarity'].get(identifier_first):
                continue

            similarity = calc_similarity(scanpath1['raw_str'], scanpath2['raw_str'])

            # Set the similarity for the first scanpath
            scanpath1['similarity'][identifier_second] = similarity

            # If the similarity object of second scanpath does not exist yet - create it
            if not scanpath2.get('similarity'):
                scanpath2['similarity'] = {}

            # Set the same similarity as above for the second scanpath
            scanpath2['similarity'][identifier_first] = similarity


def calc_similarity_to_common(scanpath_strs, scanpath_common):
    """ Calculates similarity of each individual scanpath to the common one based on the defined similarity formula. """

    # Object storing similarities of each individual scanpath to the common one
    similarity_obj = {}
    len_common = len(scanpath_common)

    # If the common scanpath is not empty (there are cases of no common scanpath sometimes)
    if len_common:
        # Calculate similarity of each scanpath to the common (trending) scanpath
        for scanpath_str in scanpath_strs:
            similarity = calc_similarity(scanpath_common, scanpath_str['raw_str'])
            similarity_obj[scanpath_str['identifier']] = similarity

    return similarity_obj


def get_most_similar_pair(scanpath_strs):
    """ Method looks for the most similar pair of scanpath strings in the given set. """

    most_similar_pair = [scanpath_strs[0], scanpath_strs[0], -1]

    for scanpath in scanpath_strs:
        for scanpath_id in scanpath['similarity']:
            if scanpath['similarity'][scanpath_id] > most_similar_pair[2]:
                most_similar_pair = [
                    scanpath,
                    get_scanpath_str_by_id(scanpath_strs, scanpath_id),
                    scanpath['similarity'][scanpath_id]
                ]

    return most_similar_pair


def convert_to_str_array(scanpaths):
    """
    Converts the dict-formatted scanpaths into an array of strings for similarity calculations. Even though an array is
    not good for searching scanpath by IDs (O(n) time), it provides an easy way to get the size, n-th element or index.
    From: [{'fixations': [['A', '150'], ['B', '750'], ['C', '300']], 'identifier': '02'}, ..]
    To: [{'raw_str': 'ABC', 'identifier': '02'}, {'raw_str': 'AC', 'identifier': '03'}, .. ]
     """

    scanpath_strs = []
    # Extract scanpaths as raw string sequences with identifiers
    for act_scanpath in scanpaths:
        act_scanpath_str = ''
        for fixation in act_scanpath['fixations']:
            act_scanpath_str += fixation[0]
        # Store the identifier and extracted string sequence in an object
        temp_scanpath = {
            'identifier': act_scanpath['identifier'],
            'raw_str': act_scanpath_str
        }
        # Push the object to the array

        scanpath_strs.append(temp_scanpath)

    return scanpath_strs




def get_scanpath_str_by_id(scanpath_strs, identifier):
    """
    Since scanpaths are stored as an array of dicts (might be changed to a dict of dicts later), sometimes
    we need this function to fetch us a particular scanpath by its id.
    """

    for scanpath in scanpath_strs:
        if scanpath['identifier'] == identifier:
            return scanpath

    raise LookupError('No such identifier in scanpath strings array: ' + identifier)


def rem_scanpath_strs_by_id(scanpath_strs, ids_to_rem):
    """
    Function removes scanpaths from the list based on provided ids. It also deletes any references to
    these scanpaths from the existing similarity objects of remaining scanpaths.
    """

    # Delete scanpaths specified by id
    for identifier in ids_to_rem:
        scanpath_strs.remove(get_scanpath_str_by_id(scanpath_strs, identifier))

    # Delete existing references to previously deleted scanpaths
    for scanpath in scanpath_strs:
        for identifier in ids_to_rem:
            if scanpath.get('similarity') and identifier in scanpath['similarity']:
                scanpath['similarity'].pop(identifier, None)


def clear_mutual_similarity(scanpath_strs):
    """ Clears the similarity object of all scanpaths in the array. """

    for scanpath in scanpath_strs:
        scanpath.pop('similarity', None)


"""
UBC Eye Movement Data Analysys Toolkit
Created on 2011-09-30

@author: skardan
"""

import math 
import skgaze.EMDAT.geometry
from warnings import warn
from skgaze.EMDAT.utils import *
from skgaze.EMDAT.Segment import *
import skgaze.EMDAT.params as params
from copy import deepcopy
from functools import reduce


class Scene(Segment):
    """A Scene is a class that represent one scene in the experiment.
    
    A Scene is a class that represent one scene in the experiment. The Scene is designed to capture all "Datapoint"s related to a target
    conceptual entity in the experiment. A Scene should have at least one Segment assigned to it. The Scene class is also used to 
    combine multiple "Segment"s and calculate the aggregated statistics for this new entity as a whole. This class is the equivalent
    of scenes as defined in Tobii studio.
    
    Attributes:
        scid: a string containing the Scene ids
        segments: a list of "Segment"s for this Scene
        
    Attributes inherited from Segment:
        alldata: A list of "Datapoint"s for this Scene
        features: A dict with feature names as its keys and feature values as its values 
        completion_time: An integer indicating total duration of the Scene in milliseconds
            minimum is 16 ms (ength of one sample with 60Hz sampling rate (ms)
        start:An integer indicating the Scene's start time in milliseconds
        end: An integer indicating the Scene's end time in milliseconds
        sample_start_ind: An integer indicating the index of the first Datapoint for this Scene in the Participant's list of all "Datapoint"s (all_data) 
        sample_end_ind: An integer indicating the index of the last Datapoint for this Scene in the Participant's list of all "Datapoint"s (all_data)
        fixation_start_ind: An integer indicating the index of the first Fixation for this Scene in the Participant's list of all "Fixation"s (fixation_data)
        fixation_end_ind: An integer indicating the index of the first Fixation for this Scene in the Participant's list of all "Fixation"s (fixation_data)
        numfixations: An integer indicating the number of "Fixation"s in this Scene
        time_gaps: a list of tuples of the form (start, end) indicating the start and end of the gaps of invalid samples in the Segement's samples
        largest_data_gap: An integer indicating the length of largest invalid gap for this Scene in milliseconds
        proportion_valid: A float indicating the proportion of valid samples over all the samples in this Scene
        proportion_valid_fix: A float indicating the proportion of (valid + restored) samples over all the samples in this Scene 
        validity1: a boolean indicating whether this Scene is valid using proportion of valid samples threshold 
        validity2: a boolean indicating whether this Scene is valid using largest acceptable gap threshold
        validity3: a boolean indicating whether this Scene is valid using proportion of (valid + restored) samples threshold
        is_valid: a boolean indicating whether this Scene is considered valid by the validity method indicated by params.VALIDITY_METHOD
        length: An integer indicating total duration of the Scene in milliseconds
        numsamples: An integer indicating total number of samples in the Scene 
        fixation_data: A list of "Fixation"s for this Scene
        fixation_start = fixation_data[0].timestamp
        fixation_end = fixation_data[-1].timestamp
        aoi_data: A list of AOI_Stat objects for relevants "AOI"s for this Scene
        has_aois: A boolean indicating if this Scene has AOI features calculated for it
        
    """

                
    def __init__(self, scid, seglist, all_data, fixation_data, Segments = None, aoilist = None,
                  prune_length= None, require_valid = True, auto_partition = False):
        """
        Args:
            scid: A string containing the id of the Scene.
            
            seglist: a list of tuples of the form (segid, start, end) defining the segments
            *Note: this method of defining segments is implemented to make batch processing of
            files defining segments easier
            
            all_data: a list of "Datapoint"s which make up this Scene.
            
            fixation_data: a list of "Fixation"s which make up this Scene.
            
            Segments: a list of "Segment"s which belong to this Scene.
                         
            aoilist: If not None, a list of "AOI"s.
            
            prune_length: If not None, an integer that specifies the time 
                interval (in ms) from the begining of each Segment of this Scene
                which samples are considered in calculations.  This can be used if, 
                for example, you only wish to consider data in the first 
                1000 ms of each Segment. In this case (prune_length = 1000),
                all data beyond the first 1000ms of the start of the "Segment"s
                will be disregarded.
            
            require_valid: a boolean determining whether invalid "Segment"s
                will be ignored when calculating the features or not. default = True 
                
            auto_partition_low_quality_segments: a boolean flag determining whether
                EMDAT should automatically split the "Segment"s which have low sample quality
                into two new ssub "Segment"s discarding the largest invalid sample gap in 
                the "Segment". default = False
        Yields:
            a Scene object
        """ 
        print()
        ########################################
        def partition_segement(new_seg, seg_start,seg_end):
            """ A helper method for splitting a Segment object into new Segments and removing gaps of invalid samples
            
            One way to deal with a low quality Segment is to find the gaps of invalid samples within its "Datapoint"s and 
            splitting the Segment into two Segments one from the beginnning of the Segment to the gap and another from after
            the gap to the end of the Segment. This can be done multiple times resulting multiple "Segment"s with higher
            quality. For example if a Segment S1 started at s1 and ended at e1 and had two invalid gaps between gs1-ge1 and 
            gs2-ge2 milliseconds, this method will generate the following three segments
                SS1: starting at s1 and ending at gs1
                SS2: starting at ge1 and ending at gs2
                SS3: starting at ge2 and ending at e1
            
            Args:
                new_seg: The Segment that is being split
                
                seg_start: An integer showing the start time of the segment in milliseconds
                
                seg_end: An integer showing the end time of the segment in milliseconds 
            
            Returns:
                subsegments: a list of newly generated "Segment"s
                
                samp_inds: a list of tuples of the form (start, end) that detrmines the index of the start and end of each 
                    new Segment in the old Segment's all_data field
                    
                fix_inds: a list of tuples of the form (start, end) that detrmines the index of the start and end of each 
                    new Segment in the old Segment's fixation_data field
            """
            timegaps = new_seg.getgaps()
            subsegments = []
            sub_segid=0
            samp_inds = []
            fix_inds = []
            last_samp_idx = 0
            last_fix_idx = 0
            sub_seg_time_start = seg_start
            for timebounds in timegaps:
                sub_seg_time_end = timebounds[0] #end of this sub_seg is start of this gap
                last_samp_idx, all_start,all_end = get_chunk(all_data, last_samp_idx, sub_seg_time_start, sub_seg_time_end)
                last_fix_idx, fix_start, fix_end = get_chunk(fixation_data, last_fix_idx, sub_seg_time_start, sub_seg_time_end)
                sub_seg_time_start = timebounds[1] #beginning of the next sub_seg is end of this gap
                if fix_end - fix_start>0:
                    try:
                        new_sub_seg = Segment(segid+"_"+str(sub_segid), all_data[all_start:all_end],
                                      fixation_data[fix_start:fix_end], aois=aoilist, prune_length=prune_length)
                    except  Exception as e:
                        warn(str(e))
                        if params.DEBUG:
                            raise
                        else:
                            continue
                else:
                    continue
                subsegments.append(new_sub_seg)
                samp_inds.append((all_start,all_end))
                fix_inds.append((fix_start, fix_end))
                sub_segid +=1
            # handling the last sub_seg
            sub_seg_time_end = seg_end #end of last sub_seg is the end of seg
            last_samp_idx, all_start,all_end = get_chunk(all_data, last_samp_idx, sub_seg_time_start, sub_seg_time_end)
            last_fix_idx, fix_start, fix_end = get_chunk(fixation_data, last_fix_idx, sub_seg_time_start, sub_seg_time_end)
            if fix_end - fix_start>0: #add the last sub_seg
                try:
                    new_sub_seg = Segment(segid, all_data[all_start:all_end],
                                      fixation_data[fix_start:fix_end], aois=aoilist, prune_length=prune_length)
                except  Exception as e:
                    warn(str(e))
                    if params.DEBUG:
                        raise
                    else:
                        new_sub_seg = None
                        
                if new_sub_seg != None:
                    subsegments.append(new_sub_seg)
                    samp_inds.append((all_start,all_end))
                    fix_inds.append((fix_start, fix_end))
            #end of handling the last sub_seg
                
            return subsegments, samp_inds, fix_inds
        ########################################
        
        if len(list(all_data))<=0:
            raise Exception('A scene with no sample data!')
        if Segments == None:
            self.segments = []
#            print "seglist",seglist
            for (segid, start, end) in seglist:
                print("segid, start, end:",segid, start, end)
                all_start, all_end = 0, len(all_data)
                fix_start, fix_end = 0, len(fixation_data)
                if fix_end - fix_start>0:
                   
                    new_seg = Segment(segid, all_data[all_start:all_end],
                                          fixation_data[fix_start:fix_end], aois=aoilist, prune_length=prune_length)
                    # except  Exception as e:
                    #     warn(str(e))
                    #     if params.DEBUG:
                    #         raise
                    #     else:
                    #         continue
                else:
                    continue
                
                if (new_seg.largest_data_gap > params.MAX_SEG_TIMEGAP) and auto_partition: #low quality segment that needs to be partitioned!
                    new_segs, samp_inds, fix_inds = partition_segement(new_seg, start, end) 
                    for nseg,samp,fix in zip(new_segs, samp_inds, fix_inds):
                            if nseg.length > params.MINSEGSIZE:
                                nseg.set_indices(samp[0],samp[1],fix[0],fix[1])
                                self.segments.append(nseg)
                else:   #good quality segment OR no auto_partition
                    new_seg.set_indices(all_start,all_end,fix_start,fix_end)
                    self.segments.append(new_seg)
        else:
            self.segments = Segments #segments are already generated
     
        self.require_valid_Segments = require_valid
        if require_valid:   #filter out the invalid Segments
            segments = list(filter(lambda x:x.is_valid,self.segments))
        else:
            segments = self.segments

        if len(segments)==0:
            raise Exception('no segments in scene %s!' %(scid))
        
        fixationlist = []
        sample_list = []
        totalfixations = 0
        firstsegtime = float('infinity')
        firstseg = None 
        for seg in segments:
            sample_st,sample_end,fix_start,fix_end = seg.get_indices()
            if params.DEBUG:
                print("sample_st,sample_end,fix_start,fix_end",sample_st,sample_end,fix_start,fix_end)
            sample_list.append(all_data[sample_st:sample_end])
            fixationlist.append(fixation_data[fix_start:fix_end])
            totalfixations += len(fixationlist[-1])
            if float(seg.start) < float(firstsegtime):
                firstsegtime = seg.start
                firstseg = seg
        
        self.firstseg = firstseg
        self.scid = scid
        self.features = {}
        self.largest_data_gap = maxfeat(self.segments,'largest_data_gap')   #self.segments is used to calculate validity of the scenes instead of segments which is only valid segments
        self.proportion_valid = weightedmeanfeat(self.segments,'numsamples','proportion_valid') #self.segments is used to calculate validity of the scenes instead of segments which is only valid segments
        self.proportion_valid_fix = weightedmeanfeat(self.segments,'numsamples','proportion_valid_fix') #self.segments is used to calculate validity of the scenes instead of segments which is only valid segments
        self.validity1 = self.calc_validity1()
        self.validity2 = self.calc_validity2()
        self.validity3 = self.calc_validity3()
        self.is_valid = self.get_validity()

        self.length = sumfeat(segments,'length')
        if self.length == 0:
            raise Exception('Zero length segments!')
        self.features['numsegments'] = len(segments)
        self.features['length'] = self.length
        self.start = minfeat(segments,'start')
        self.numfixations = sumfeat(segments,'numfixations')
        self.end = maxfeat(segments,'end')
        self.numsamples = sumfeat(segments, 'numsamples')
        self.features['numsamples'] = self.numsamples
        
        self.numfixations = sumfeat(segments, 'numfixations')
        self.features['numfixations'] = self.numfixations
        if prune_length == None:
            if self.numfixations != totalfixations:
                raise Exception('error in fixation count for scene:'+self.scid)
            #warn ('error in fixation count for scene:'+self.scid)
        self.features['fixationrate'] = float(self.numfixations) / self.length
        if self.numfixations > 0:
            self.features['meanfixationduration'] = weightedmeanfeat(segments,'numfixations',"features['meanfixationduration']")
            self.features['stddevfixationduration'] = stddev(list(map(lambda x: float(x.fixationduration), reduce(lambda x,y: x+y ,fixationlist))))##
            self.features['sumfixationduration'] = sumfeat(segments, "features['sumfixationduration']")
            self.features['fixationrate'] = float(self.numfixations)/self.length
            distances = self.calc_distances(fixationlist)
            abs_angles = self.calc_abs_angles(fixationlist)
            rel_angles = self.calc_rel_angles(fixationlist)
        else:
            self.features['meanfixationduration'] = 0
            self.features['stddevfixationduration'] = 0
            self.features['sumfixationduration'] = 0
            self.features['fixationrate'] = 0
            distances = []
        if len(distances) > 0:
            self.features['meanpathdistance'] = mean(distances)
            self.features['sumpathdistance'] = sum(distances)
            self.features['stddevpathdistance'] = stddev(distances)
            self.features['eyemovementvelocity'] = self.features['sumpathdistance']/self.length
            self.features['sumabspathangles'] = sum(abs_angles)
            self.features['meanabspathangles'] = mean(abs_angles)
            self.features['stddevabspathangles'] = stddev(abs_angles)
            self.features['sumrelpathangles'] = sum(rel_angles)
            self.features['meanrelpathangles'] = mean(rel_angles)
            self.features['stddevrelpathangles'] = stddev(rel_angles)
        else:
            self.features['meanpathdistance'] = 0
            self.features['sumpathdistance'] = 0
            self.features['stddevpathdistance'] = 0
            self.features['sumabspathangles'] = 0
            self.features['meanabspathangles']= 0
            self.features['stddevabspathangles']= 0
            self.features['sumrelpathangles'] = 0
            self.features['meanrelpathangles']= 0
            self.features['stddevrelpathangles'] = 0
        self.has_aois = False
        if aoilist:
            self.set_aois(segments, aoilist)
            
    def getid(self):
        """Returns the scid for this Scene
        
        Returns: a string conataining the scid for this Scene
        """
        return self.scid
    
    def set_aois(self, segments, aois):
        """Sets the "AOI"s relevant to this Scene
        
        Args:
            segments: a list of "Segment"s which belong to this Scene.

            aois: a list of "AOI"s relevant to this Scene        
        """
        if len(aois) == 0:
            print("no AOI:",self.segid)
        self.aoi_data={}
        for seg in segments:
            for aid in seg.aoi_data.keys():
                if seg.aoi_data[aid].isActive:
                    if aid in self.aoi_data:
                        self.aoi_data[aid] = merge_aoistats(self.aoi_data[aid],seg.aoi_data[aid], self.features['length'],self.numfixations)
                    else:
                        self.aoi_data[aid] = deepcopy(seg.aoi_data[aid])
        
        firstsegaois = self.firstseg.aoi_data.keys()            
        for aid in self.aoi_data.keys():
            if aid in firstsegaois:
                self.aoi_data[aid].features['timetofirstfixation'] = deepcopy(self.firstseg.aoi_data[aid].features['timetofirstfixation'])
            else:
                self.aoi_data[aid].features['timetofirstfixation'] = float('inf')
                
        #maois.features['averagetimetofirstfixation'] = ?
        #maois.features['averagettimetolastfixation'] = ?
            self.has_aois = True

    def calc_distances(self, fixdatalists):
        """returns the Euclidean distances between a sequence of "Fixation"s
    
        Args:
            fixdatalists: a list of lists of "Fixation"s
        """
        distances = []
        for fixdata in fixdatalists:
            #print "fixdata", fixdata
            if not(fixdata):
                continue
            lastx = fixdata[0].mappedfixationpointx
            lasty = fixdata[0].mappedfixationpointy
    
            for i in range(1, len(fixdata)):
                x = fixdata[i].mappedfixationpointx
                y = fixdata[i].mappedfixationpointy
                dist = math.sqrt((x-lastx)**2 + (y-lasty)**2)
                distances.append(dist)
                lastx = x
                lasty = y
        return distances
    
    def calc_abs_angles(self, fixdatalists):
        """returns the absolute angles between a sequence of "Fixation"s that build a scan path.
        
        Abosolute angle for each saccade is the angle between that saccade and the horizental axis
    
        Args:
            fixdatalists: a list of lists of "Fixation"s
            
        Returns:
            a list of absolute angles for the saccades formed by the given sequence of "Fixation"s in Radiant
        """
        abs_angles = []
        for fixdata in fixdatalists:
            if not(fixdata):
                continue
            lastx = fixdata[0].mappedfixationpointx
            lasty = fixdata[0].mappedfixationpointy
    
            for i in range(1,len(fixdata)):
                x = fixdata[i].mappedfixationpointx
                y = fixdata[i].mappedfixationpointy
                (_, theta) = geometry.vector_difference((lastx,lasty), (x, y))
                abs_angles.append(abs(theta))
                lastx=x
                lasty=y

        return abs_angles

    def calc_rel_angles(self, fixdatalists):
        """returns the relative angles between a sequence of "Fixation"s that build a scan path in Radiant
        
        Relative angle for each saccade is the angle between that saccade and the previous saccade.
    
        Args:
            fixdatalists: a list of lists of  "Fixation"s
            
        Returns:
            a list of relative angles for the saccades formed by the given sequence of "Fixation"s in Radiant
        """
        rel_angles = []
        
        for fixdata in fixdatalists:
            if not(fixdata):
                continue
            lastx = fixdata[0].mappedfixationpointx
            lasty = fixdata[0].mappedfixationpointy
    
            for i in range(1, len(fixdata)-1):
                x = fixdata[i].mappedfixationpointx
                y = fixdata[i].mappedfixationpointy
                nextx = fixdata[i+1].mappedfixationpointx
                nexty = fixdata[i+1].mappedfixationpointy
                (_, theta) = geometry.vector_difference((x,y), (lastx, lasty))
                (_, nextheta) = geometry.vector_difference((x,y), (nextx, nexty))
                theta = abs(theta-nextheta)
                rel_angles.append(theta)
                lastx=x
                lasty=y

        return rel_angles   

def merge_aoistats(main_AOI_Stat,new_AOI_Stat,total_time,total_numfixations):
        """a helper method that updates the AOI_Stat object of this Scene with a new AOI_Stat object
        
        Args:
            main_AOI_Stat: AOI_Stat object of this Scene
            
            new_AOI_Stat: a new AOI_Stat object
            
            total_time:
            
            total_numfixations:
        
        Returns:
            the updated AOI_Sata object
        """   
        maois = main_AOI_Stat
        maois.features['numfixations'] += new_AOI_Stat.features['numfixations']
        maois.features['longestfixation'] = max(maois.features['longestfixation'],new_AOI_Stat.features['longestfixation'])
        maois.features['totaltimespent'] += + new_AOI_Stat.features['totaltimespent'] 
        maois.features['proportiontime'] = float(maois.features['totaltimespent'])/total_time
        maois.features['proportionnum'] = float(maois.features['numfixations'])/total_numfixations
        if maois.features['totaltimespent']>0: 
            maois.features['fixationrate'] = float(maois.features['numfixations'])/maois.features['totaltimespent']
        else:
            maois.features['fixationrate'] = 0.0
                
                
            
        #calculating the transitions to and from this AOI and other active AOIs at the moment
        new_AOI_Stat_transition_aois = filter(lambda x: x.startswith(('numtransto_','numtransfrom_')),new_AOI_Stat.features.keys())
        if params.DEBUG:
            print("segement's transition_aois",new_AOI_Stat_transition_aois)
            
        maois.total_tans_to += new_AOI_Stat.total_tans_to       #updating the total number of transition to this AOI
        maois.total_tans_from += new_AOI_Stat.total_tans_from   #updating the total number of transition from this AOI
       
        for feat in new_AOI_Stat_transition_aois:
            if feat in maois.features.copy():
                maois.features[feat] += new_AOI_Stat.features[feat]
            else:
                maois.features[feat] = new_AOI_Stat.features[feat]
#            if feat.startswith('numtransto_'):
#                sumtransto += maois.features[feat]
#            else:
#                sumtransfrom += maois.features[feat]
    
    
# updating the proportion tansition features based on new transitions to and from this AOI        
        maois_transition_aois = list(filter(lambda x: x.startswith(('numtransto_','numtransfrom_')),maois.features.keys())) #all the transition features for this AOI should be aupdated even if they are not active for this segment
        for feat in maois_transition_aois.copy():
            if feat.startswith('numtransto_'):
                aid = feat.lstrip('numtransto_')
                if maois.total_tans_to > 0:

                    maois.features['proptransto_%s'%(aid)] = float(maois.features[feat]) / maois.total_tans_to
                else:
                    maois.features['proptransto_%s'%(aid)] = 0
            else:
                aid = feat.lstrip('numtransfrom_')
                if maois.total_tans_from > 0:
                    
                    maois.features['proptransfrom_%s'%(aid)] = float(maois.features[feat]) / maois.total_tans_from
                else:
                    maois.features['proptransfrom_%s'%(aid)] = 0
        ###endof trnsition calculation
        return maois

def weightedmeanfeat(obj_list, totalfeat,ratefeat):
    """a helper method that calculates the weighted average of a target feature over a list of Segments
    
    Args:
        obj_list: a list of Segments which all have a numeric field for which the weighted average is calculated 
        
        totalfeat: a string containing the name of the feature that has the total value of the target feature 
        
        ratefeat: a string containing the name of the feature that has the rate value of the target feature
    
    Returns:
        the weighted average of the ratefeat over the Segments
    """
    num_valid = float(0)
    num = 0

    for obj in obj_list:
        t = eval('obj.'+totalfeat)
        num_valid += t * eval('obj.'+ratefeat)
        num += t
    if num != 0:
        return num_valid / num
    return 0
    

def sumfeat(obj_list, feat):
    """a helper method that calculates the sum of a target feature over a list of objects
    
    Args:
    
        obj_list: a list of objects
        
        feat: a string containing the name of the target feature
    
    Returns:
        the sum of the target feature over the given list of objects
    """
    sum = 0
    for obj in obj_list:
        sum += eval('obj.'+feat)
    return sum

def minfeat(obj_list, feat):
    """a helper method that calculates the min of a target feature over a list of objects
    
    Args:
    
        obj_list: a list of objects
        
        feat: a string containing the name of the target feature
    
    Returns:
        the min of the target feature over the given list of objects
    """
    min = float('+infinity')
    for obj in obj_list:
        val = eval('obj.'+feat)
        if min > float(val):
            min = float(val)
    return min     
    
def maxfeat(obj_list, feat):
    """a helper method that calculates the max of a target feature over a list of objects
    
    Args:
    
        obj_list: a list of objects
        
        feat: a string containing the name of the target feature
    
    Returns:
        the max of the target feature over the given list of objects
    """
    max = float('-infinity')
    for obj in obj_list:
        val = eval('obj.'+feat)
        if max < float(val):
            max = float(val)
    return max          



from skgaze.EMDAT.data_structures import Datapoint
from skgaze.EMDAT.BasicParticipant import BasicParticipant

class EMDAT:

    def __init__(self, dataset,pid=False):
        self.pid = pid
        self.dataset = dataset
       
        

    def createInstance(self):
        result = {}
        if self.pid:
           for participant in self.dataset.participants:
               if participant.id == self.pid:   
                    return BasicParticipant(self.pid,False,self.dataset.participants[0].fixations,self.dataset.participants[0].fixations,self.dataset.participants[0],aoifile=self.dataset.participants[0].aoi_array)
        for participant in self.dataset.participants:
            result[participant.id] = BasicParticipant(participant.id,False,participant.fixations,participant.fixations,participant,aoifile=participant.aoi_array)
        return result
        


"""
UBC Eye Movement Data Analysys Toolkit
Created on 2011-08-25

@author: skardan

Commonly used helper methods
"""
from skgaze.EMDAT.data_structures import Fixation
import skgaze.EMDAT.params as params
import math


def point_inside_polygon(x,y,poly):
    """Determines if a point is inside a given polygon or not
    
        The algorithm is called "Ray Casting Method".
        
    Args:
        poly: is a list of (x,y) pairs defining the polgon 
        
    Returns:
        True or False.
    """
    n = len(poly)
    
    if n==0:
        return False
    
    inside = False

    p1x,p1y = poly[0]
    for i in range(n+1):
        p2x,p2y = poly[i % n]
        if y > min(p1y,p2y):
            if y <= max(p1y,p2y):
                if x <= max(p1x,p2x):
                    if p1y != p2y:
                        xinters = (y-p1y)*(p2x-p1x)/(p2y-p1y)+p1x
                    if p1x == p2x or x <= xinters:
                        inside = not inside
        p1x,p1y = p2x,p2y

    return inside   

def get_chunk(data, ind, start, end):
    """Returns index of first and last records in data that fall within a time interval (start-end) 
    Args:
        data: a list of subsequent Fixations or Datapoints
        ind: an integer indicating the starting index in data for search, if not known 
            should be set to zero.
        start: an integer indicating the start of interval in milliseconds
        end: an integer indicating the end of interval in milliseconds
        
    Returns:
        curr_ind: an integer indicating the index of the next record for search. 
            This is useful if you are performing a series searches in a sequential manner.
            The method can start the next search from this index instead of beginning of the list.               
        start_ind: an integer indicating the index of first record in the list that falls within
            the given time interval
        end_ind: an integer indicating the index of last record in the list that falls within
            the given time interval
    """
    datalen = len(data)
    curr_ind = ind
    if int(curr_ind) < datalen:
        if isinstance(data[curr_ind],Fixation): #if it is a fixation
            if params.INCLUDE_HALF_FIXATIONS: 
                while curr_ind < datalen and data[curr_ind].timestamp < start:
                    curr_ind += 1
            
                if data[curr_ind-1].fixationduration!= None: # if the last fixation before, is mostly in this segment
                    if (data[curr_ind-1].timestamp + (data[curr_ind-1].fixationduration)/2.0) > start:
                        curr_ind -=1
                start_ind = curr_ind
                while curr_ind < datalen and (data[curr_ind].timestamp + (data[curr_ind].fixationduration)) <= end:
                    curr_ind += 1
            
                 
                if curr_ind == start_ind:   # an empty chunk!
                    end_ind = curr_ind -1
                elif data[curr_ind-1].fixationduration!= None: # if the last fixation is mostly outside this segment
                    if (data[curr_ind-1].timestamp + (data[curr_ind-1].fixationduration)/2.0) > end:
                        end_ind = curr_ind - 2 
                    else: 
                        end_ind = curr_ind -1
                else:
                    end_ind = curr_ind -1
            else:   #No half fixation inclusion
                while curr_ind < datalen and data[curr_ind].timestamp < start:
                    curr_ind += 1
            
                start_ind = curr_ind
                while curr_ind < datalen and (data[curr_ind].timestamp + (data[curr_ind].fixationduration)) <= end:
                    curr_ind += 1
            
                 
                if curr_ind == start_ind:   # a no point chunk!
                    end_ind = curr_ind -1
                elif data[curr_ind-1].fixationduration!= None: # if the last fixation is mostly outside this segment
                    if (data[curr_ind-1].timestamp + (data[curr_ind-1].fixationduration)/2.0) > end:
                        end_ind = curr_ind - 2 
                    else: 
                        end_ind = curr_ind -1
                else:
                    end_ind = curr_ind -1            
        else: # if this is not a Fixation we do not have to worry about half fixations
            while int(curr_ind) < datalen and int(data[curr_ind].timestamp) < start:
                curr_ind += 1
        
            start_ind = curr_ind
            while curr_ind < datalen and data[curr_ind].timestamp <= end:
                curr_ind += 1
        
            end_ind = curr_ind -1
        
        end_ind +=1 #because the last index is not inclusive in Python!
    else:
        curr_ind = start_ind = end_ind = datalen
        
    return curr_ind, start_ind, end_ind

def stddev(data):
    """Returns the standard deviation of a list of numbers
    
    Args:
        data: a list of numbers
    
    returns:
        a float that is the std deviation of the list of numbers or NAN if it is undefined
    """
    if len(data)< 2:
        return float('nan')
    m = mean(data)
    return math.sqrt(sum(map(lambda x: (x-m)**2, data))/float(len(data)-1))
    
def mean(data):
    """Returns the average of a list of numbers
    
    Args:
        data: a list of numbers
    
    returns:
        a float that is the average of the list of numbers
    """
    if len(data)==0:
        return 0
    return sum(data) / float(len(data))


def get_longest_common_substring(s1, s2):
    # Source: https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Longest_common_substring

    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]
    longest, x_longest = 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    x_longest = x
            else:
                m[x][y] = 0
    return s1[x_longest - longest: x_longest]


from collections import defaultdict
from skgaze.core.Dataset import Dataset
import os
import math

class SPAM:


    def __init__(self, dataset):
        self.raw_sequences = dataset.create_raw_sequences()
        self.listOfAois = self.simplify_aoi_array(dataset.aoi_array)
        self.listOfScanpaths = self.create_list_of_scanpaths(dataset)
        self.result = []

   
    def simplify_aoi_array(self, aoi_array):
        spam_aoi_array = []
        for aoi in aoi_array:
        	spam_aoi_array.append(aoi.aoi_char)
        return spam_aoi_array

    def create_list_of_scanpaths(self,dataset):
        result = []
        for participant in dataset.participants:
            result.append(participant.scanpath)
        return result


    def runSpam(self, minsup_percenta=0.5, minlength=1, maxlength=4, maxgap=1):
    
        dlzkasekvencii = []
        dlzkasuboru = 0
        SPAM.pocet_najdenych_vzorov = 0

        supportnumb = 0
        sid = 0  # id participanta
        tid = 0  # id AOI
        verticalDB = {}
        supportDB = {}
       
        # Krok 1 ulozime si poziciu prvej AOI kazdej sekvencie pohladu, a zaroven aj poslednu poslednej AOI
    
        for line in range(0,len(self.listOfScanpaths)):
            dlzkasekvencii.append(dlzkasuboru)
            for i in self.listOfScanpaths[line]:
                dlzkasuboru += 1
            #dlzkasekvencii.append(dlzkasuboru)
            #for ch in line:
            #    if ((ch >= 'A') and (ch <= 'Z')):
            #        dlzkasuboru += 1
        dlzkasekvencii.append(dlzkasuboru)  # ulozime si aj poziciu posledneho bitu
  


        #Krok 2 vypocitame si minsup
        minsup = math.ceil(float(minsup_percenta) * (len(dlzkasekvencii)-1))
        if minsup == 0:
            minsup = 1

        #Krok 3
        #vytvorime vertikalnu reprezentaciu databazy
        #ukladame poziciu kazdej AOI a poradie sekvencie kazdej AOI priradime
        #bitmapu (ak taka AOI este neexistuje, vytvorime ju a vlozime do vertikalnej reprezentacie databazy)
        #nasledne musime nastavit bit na 1 na takej pozicii v bitmape, v ktorej sa AOI vyskytuje

  
        for line in range(0,len(self.listOfScanpaths)):
            for i in self.listOfScanpaths[line]:
                if i[0] not in verticalDB:
                    
                    verticalDB[i[0]] = [0] * dlzkasekvencii[-1]
                    supportDB[i[0]] = 0
                verticalDB[i[0]][dlzkasekvencii[sid] + tid] = 1
                tid += 1
            tid = 0
            sid += 1    #posun na dalsi riadok/participanta
        sid -= 1
        #{'A': [1, 0, 0, 1, 0, 0...], 'B': [0, 1, 0, 0, 0, 0...], 'C': [0, 0, 1, 0, 1, 1...]}

        #Krok 4 zistime support jednotlivych oblasti zaujmu
        for i in verticalDB:
            for j in range(sid+1):
                for k in range(0, (dlzkasekvencii[j+1] - dlzkasekvencii[j])):
                    if verticalDB[i][dlzkasekvencii[j]+k] is 1:
                        supportnumb += 1
                        supportDB[i] = supportnumb
                        break
            supportnumb = 0

        #Krok 5
        #Ulozime si vzory, ktore sa nachadzaju vo viacerych sekvenciach ako je minsup
        frequentItems = []
        
        for i in supportDB:
            if int(supportDB.get(i)) >= int(minsup):
                if (int(minlength) <= 1) and (int(maxlength) >= 1):
                    #print('[\'',i,'\']', "support", supportDB[i])
                    SPAM.pocet_najdenych_vzorov += 1
                    self.result.append([SPAM.pocet_najdenych_vzorov, i, supportDB[i]])
                frequentItems.append(i)

        if (maxlength == 1):
            return
        #Krok 6
        #vyhladavanie do hlbky s vykonanim S-Step
        for i in verticalDB:
            prefix = []
            prefix.append(i)
            self.dfsPruning(prefix, verticalDB[i], frequentItems, 2, minsup, minlength, maxlength, dlzkasekvencii,
                             maxgap, verticalDB)

        return self.result
    supportWithoutGapTotal = 0
    

    #s_n = iba vzory, ktore vieme, ze sa budu urcite vyskytovat v najcastejsich vzoroch = frequentitems
    def dfsPruning(self,prefix, bitmap_prefix, s_n, m, minsup, minlength, maxlength, dlzkasekvencii,
                   maxgap, verticalDB):
        sTemp = []
        sTempBitmaps = {}
        supportDB2 = {}     ##ukladame si support prvkov

        for i in s_n:
            newBitmap = self.newBitmap(bitmap_prefix, dlzkasekvencii, maxgap, verticalDB[i])
            if (SPAM.supportWithoutGapTotal >= minsup):
                sTemp.append(i)
                sTempBitmaps[i] = newBitmap
                supportDB2[i] = SPAM.supportOfItem

            SPAM.supportWithoutGapTotal = 0
            SPAM.supportOfItem = 0

        for j in range(0, len(sTemp)):
            item = sTemp[j]

            prefixSStep = self.cloneSequence(prefix)
            prefixSStep.append(item)

            newBitmap = sTempBitmaps[sTemp[j]]
            if (supportDB2[item] >= minsup):
                if (m >= int(minlength)):
                    #print(prefixSStep, "support", supportDB2[item])
                    SPAM.pocet_najdenych_vzorov += 1
                    self.result.append([SPAM.pocet_najdenych_vzorov, prefixSStep, supportDB2[item]])
                if (int(maxlength) > m):
                    self.dfsPruning(prefixSStep, newBitmap, sTemp, m+1, minsup, minlength, maxlength, dlzkasekvencii,
                                    maxgap, verticalDB)

    #global parameters
    lastSID = -1
    supportOfItem = 0
    # SavePaterns
    pocet_najdenych_vzorov = 0
    najdeneVzory = {}

    def cloneSequence(self, items):
        sequence = []
        for i in items:
            sequence.append(i)
        return sequence

    #bitmapa_pismena = toho pismena, ktore sa vola v kroku 6; pri rekurzivnom volani je to novovzniknuta bitmapa
    #pismeno = bitmapa toho pismena, ktore sa prave nachadza v cykle frekventovanych AOI

    def newBitmap(self, bitmap_prefix, dlzkasekvencii, maxgap, pismeno):
        newBitmap = [0] * dlzkasekvencii[-1]
        #sid poslednej sekvencie, vlozenej do bitmapy, ktory obsahuje 1
        SPAM.lastSID = -1
        k = 0
        premenna = True
        #ak je maxgap 0, tak hladame spolocne vzory, ktore mozu mat medzi sebou lubovolne velku vzdialenost
        if int(maxgap) == 0:
            bitK = bitmap_prefix.index(1, k, dlzkasekvencii[-1])
            while True:
                sid = self.bitToSid(bitK, dlzkasekvencii)
                lastBitofSid = dlzkasekvencii.__getitem__(sid+1) - 1
                match = False
                if 1 not in pismeno[bitK+1:]:
                    premenna = False
                else:
                    bit = pismeno.index(1, bitK + 1, dlzkasekvencii[-1])

                while premenna and (bit <= lastBitofSid):
                    newBitmap[bit] = 1
                    match = True
                    if (1 not in pismeno[bit+1:]):
                        break
                    bit = pismeno.index(1, bit + 1, dlzkasekvencii[-1])

                if match == True:
                    if (sid != SPAM.lastSID):
                        SPAM.supportWithoutGapTotal += 1
                        SPAM.supportOfItem += 1
                        SPAM.lastSID = sid

                k = bitK + 1
                if 1 not in bitmap_prefix[k:]:
                    break
                else:
                    bitK = bitmap_prefix.index(1, k, dlzkasekvencii[-1])

        #ak je gap medzi AOI ziadna a vacsia
        else:
            previousSid = -1
            bitK = bitmap_prefix.index(1, k, dlzkasekvencii[-1])

            #AND medzi bitmapy (bitmap_prefix a pimeno)
            while True:
                sid = self.bitToSid(bitK, dlzkasekvencii)
                lastBitofSid = dlzkasekvencii.__getitem__(sid + 1) - 1
                match = False
                matchWithoutGap = False

                if 1 not in pismeno[bitK+1:]:
                    premenna = False
                else:
                    bit = pismeno.index(1, bitK + 1, dlzkasekvencii[-1])

                #jedtnotkovy bit sa nachadza este stale v tej istej sid (sekvencie pohladu),
                #ak nie, preskocim a idem na dalsiu sekvenciu
                while premenna and (bit <= lastBitofSid):
                    matchWithoutGap = True

                    if (bit - bitK > int(maxgap)):
                        break

                    newBitmap[bit] = 1
                    match = True
                    if (1 not in pismeno[bit+1:]):
                        break
                    #vyuzitie pri ak je maxgap > 1
                    bit = pismeno.index(1, bit + 1, dlzkasekvencii[-1])

                if matchWithoutGap and previousSid!= sid:
                    SPAM.supportWithoutGapTotal += 1
                    previousSid = sid

                if match == True:
                    #++support ak je to iny participant
                    if (sid != SPAM.lastSID):
                        SPAM.supportOfItem += 1
                    SPAM.lastSID = sid

                #Vyhladanie dalsieho najblizsie bitu v bitmape, nastaveneho na 1, ak taky neexistuje, koncime
                k = bitK + 1
                if 1 not in bitmap_prefix[k:]:
                    break
                else:
                    bitK = bitmap_prefix.index(1, k, dlzkasekvencii[-1])

        return newBitmap

    #SavePaterns
    pocet_najdenych_vzorov = 0
    najdeneVzory = {}

    def bitToSid(self, bitK, dlzkasekvencii):
        for i in range(1, len(dlzkasekvencii)):
            if bitK < dlzkasekvencii[i]:
                return i-1
        return i

    #Save patterns of length 1 to file


'''
Created on 2012-08-23

@author: skardan
'''
from BasicParticipant import *
from Participant import export_features_all, write_features_tsv
from ValidityProcessing import output_Validity_info_Segments, output_percent_discarded, output_Validity_info_Participants

ul =        [61, 62]    # list of user recordings (files extracted for one participant from Tobii studio)
uids =      [61, 62]    # User ID that is used in the external logs (can be different from above but there should be a 1-1 mapping)

alogoffset =[ 3,  2]    # the time sifference between the eye tracker logs and the external log

####### Testing error handling
#ul =        [61, 62, 63]    # list of user recordings (files extracted for one participant from Tobii studio)
#uids =      [61, 62, 63]    # User ID that is used in the external logs (can be different from above but there should be a 1-1 mapping)
#
#alogoffset =[ 3,  2, 2]    # the time sifference between the eye tracker logs and the external log



###### Read participants
ps = read_participants_Basic(user_list = ul,pids = uids, log_time_offsets = alogoffset, datadir=params.EYELOGDATAFOLDER, 
                           prune_length = None, 
                           aoifile = "./sampledata/general.aoi",
#                           aoifile = "./sampledata/Dynamic_1.aoi",
                           require_valid_segs = False, auto_partition_low_quality_segments = True)
print
######

#explore_validation_threshold_segments(ps, auto_partition_low_quality_segments = False)
output_Validity_info_Segments(ps, auto_partition_low_quality_segments_flag = False, validity_method = 3)
output_percent_discarded(ps,'./outputfolder/disc.csv')
output_Validity_info_Segments(ps, auto_partition_low_quality_segments_flag = False, validity_method = 2, threshold_gaps_list = [100, 200, 250, 300],output_file = "./outputfolder/Seg_val.csv")
output_Validity_info_Participants(ps, include_restored_samples =True, auto_partition_low_quality_segments_flag = False)


##### WRITE features to file
print
aoi_feat_names = (map(lambda x:"Test"+'_'+x, params.aoigeneralfeat))
print "exporting:", params.featurelist, "\n", aoi_feat_names
write_features_tsv(ps, './outputfolder/smaple_features.tsv',featurelist = params.featurelist, aoifeaturelabels=aoi_feat_names, id_prefix = False)



# pylint: skip-file
"""
UBC Eye Movement Data Analysis Toolkit
Generic Participant Class
Created on 2011-09-25

@author: skardan
"""
import string
from skgaze.EMDAT.data_structures import *
import skgaze.EMDAT.params as params
from skgaze.EMDAT.Scene import Scene
import skgaze.EMDAT.Recording as Recording



class Participant():
    """
    A class that holds the information for one Participant in the experiment
    """

    def __init__(self, pid, eventfile, datafile, fixfile, segfile, log_time_offset = None, aoifile = None, prune_length= None, 
                 require_valid_segs = True, auto_partition_low_quality_segments = False):
        """Inits BasicParticipant class
        Args:
            pid: Participant id
            
            eventfile: a string containing the name of the "Event-Data.tsv" file for this participant
            
            datafile: a string containing the name of the "all-Data.tsv" file for this participant
            
            fixfile: a string containing the name of the "Fixation-Data.tsv" file for this participant
            
            segfile: a string containing the name of the '.seg' file for this participant
            
            log_time_offset: If not None, an integer indicating the time offset between the 
                external log file and eye tracking logs
            
            aoifile: If not None, a string conatining the name of the '.aoi' file 
                with definitions of the "AOI"s.
            
            prune_length: If not None, an integer that specifies the time 
                interval (in ms) from the begining of each Segment in which
                samples are considered in calculations.  This can be used if, 
                for example, you only wish to consider data in the first 
                1000 ms of each Segment. In this case (prune_length = 1000),
                all data beyond the first 1000ms of the start of the "Segment"s
                will be disregarded.
                
            require_valid_segs: a boolean determining whether invalid "Segment"s
                will be ignored when calculating the features or not. default = True 
                
            auto_partition_low_quality_segments: a boolean indicating whether EMDAT should 
                split the "Segment"s which have low sample quality, into two new 
                sub "Segment"s discarding the largest gap of invalid samples.
            
        Yields:
            a Participant object
        """
        self.pid = pid
        self.require_valid_segments = require_valid_segs

    def export_features(self, featurelist=None, aoifeaturelist=None, aoifeaturelabels = None,
                        id_prefix = False, require_valid = True):
        """Returns feature names and their values for this Participant
        
        Args:
            featurelist: if not None, a list of strings containing the name of the features to be
                returned
            
            aoifeaturelist: if not None, a list of features to be returned for each of the "AOI"s. 
            aoifeaturelabels: if not None, a list of AOI related features to be returned.
            *Note: while aoifeaturelist is a subset of features that will be returned for all 
            "AOI"s, aoifeaturelabels contains the exact AOI feature name, i.e., a feature
            of the form: [AOI name]_[feature name]
            For example for an AOI called 'graph', aoifeaturelabels may contain 'graph_fixationrate'  

            id_prefix: a boolean determining if the method should also export the participant id 
            
            require_valid: a boolean determining if only valid segments should be used when
            calculating the features. default = True
            
        Returns:
            featnames: a list of feature names sorted alphabetically
            featvals: a corresponding list of feature values
            e.g.
            featnames = ['fixationrate', 'length', 'meanabspathangles']
            featvals  = [0.00268522882294', '1529851', '1.60354714212']
        """
        data = []
        featnames = []
        if id_prefix:
            featnames.append('Part_id')
        featnames.append('Sc_id')
        first = True
        for sc in self.scenes:
            if not sc.is_valid and require_valid:
                print("User %s:Scene %s dropped because of 'require_valid'" %(self.id,sc.scid))
                continue
            sc_feats = []
            if id_prefix:
                sc_feats.append(self.id)
            sc_feats.append(sc.scid)
            fnames, fvals = sc.get_features(featurelist = featurelist,
                                           aoifeaturelist = aoifeaturelist, 
                                           aoifeaturelabels = aoifeaturelabels)
            if first: featnames += fnames
            sc_feats += fvals
            first = False
            data.append(sc_feats)            

        return featnames, data

    def export_features_tsv(self, featurelist=None, aoifeaturelist=None, id_prefix = False, 
                            require_valid = True):
        """Returns feature names and their values for this Participant in a tab separated format
        
        Args:
            featurelist: if not None, a list of strings containing the name of the features to be
                returned
            
            aoifeaturelist: if not None, a list of features to be returned for each of the "AOI"s. 

            id_prefix: a boolean determining if the method should also export the participant id.
             
            
            require_valid: a boolean determining if only valid segments should be used when
            calculating the features. default = True
            
        Returns:
            A two-line string with the first line having the feature names sorted alphabetically
            and separated by a tab '/t', and the second line containing the corresponding values
            separated by a tab '/t'
            For example:
            fixationrate    length    meanabspathangles
            0.00268522882294    1529851    1.60354714212
        """
        featnames, data  = self.export_features(featurelist, aoifeaturelist = aoifeaturelist, 
                                                id_prefix = id_prefix, require_valid = require_valid)
        ret = ''
        for t in featnames:
            ret += t + '\t'
        ret += '\n'
        for t in data:
            for item in t:
                ret += str(item) + '\t'
            ret += '\n'
        return ret
    
    def print_(self):
        """Outputs all feature names and their values for this Participant to the console        
        """       
        def format_list(list,leng=None):
            """
            """
            out=''
            if leng == None:
                maxlen=0
                for j in list:
                    st = repr(j)
                    if len(st) > maxlen:
                        maxlen = len(st)
                for j in list:
                    out+= repr(j).rjust(maxlen+1)
                return out,maxlen+1
            else:
                for j in list:
                    st = repr(j)
                    out+= st.rjust(leng)
                return out,leng
 
        print("PID:",self.id)
        
        for seg in self.segments:
            featnames = []
            if not seg.is_valid:
                continue
            seg_feats = []
            featnames.append('seg_id')
            seg_feats.append(seg.segid)
            fnames, fvals = seg.get_features()
            featnames += fnames
            seg_feats += fvals
            o,l= format_list(featnames)
            print(o)
            print(format_list(seg_feats,l))
            
        for sc in self.scenes:
            featnames = []
            if not sc.is_valid:
                continue
            sc_feats = []
            featnames.append('sc_id')
            sc_feats.append(sc.scid)
            fnames, fvals = sc.get_features()
            featnames += fnames
            sc_feats += fvals
            o,l= format_list(featnames)
            print(o)
            print(format_list(sc_feats,l))
    

def read_participants(segsdir, datadir, prune_length = None, aoifile = None):
    """Placeholder for a method that generates Participant objects for each participant 
    in the experiment 
    """
    participants = []
    raise Exception("You should override the default Participant.read_participants method!")
    return participants

def export_features_all(participants, featurelist = None, aoifeaturelist = None, aoifeaturelabels=None,
                         id_prefix = False, require_valid = True):
    """Returns feature names and their values for a list of "Participant"s
    
    Args:
        participants: a list of "Participant"s
        featurelist: if not None, a list of strings containing the name of the features to be
            returned
        
        aoifeaturelist: if not None, a list of features to be returned for each of the "AOI"s. 
        aoifeaturelabels: if not None, a list of AOI related features to be returned.
        *Note: while aoifeaturelist is a subset of features that will be returned for all 
        "AOI"s, aoifeaturelabels contains the exact AOI feature name, i.e., a feature
        of the form: [AOI name]_[feature name]
        For example for an AOI called 'graph', aoifeaturelabels may contain 'graph_fixationrate'  

        id_prefix: a boolean determining if the method should also export the participant id 
        
        require_valid: a boolean determining if only valid segments should be used when
        calculating the features. default = True
        
    Returns:
        featnames: a list of feature names sorted alphabetically
        featvals: a corrsponding list of feature values
        e.g.
        featnames = ['fixationrate', 'length', 'meanabspathangles']
        featvals  = [0.00268522882294', '1529851', '1.60354714212']
    """
    data = []
    featnames = []
    if participants:
        for p in participants:
            if not(p.is_valid()):
                print("user",p.id,"was not valid")
                continue
            fnames, fvals = p.export_features(featurelist=featurelist, aoifeaturelist=aoifeaturelist, 
                                              aoifeaturelabels = aoifeaturelabels,
                                              id_prefix=id_prefix, require_valid = require_valid)
            featnames = fnames
            data += fvals
    else:
        raise NameError('No participants were passed to the function')
    
    return featnames, data

def write_features_tsv(participants, outfile, featurelist = None, aoifeaturelist =  None, 
                       aoifeaturelabels=None, id_prefix = False):
    """Returns feature names and their values for a list of "Participant"s in a tsv-format file
    
    This method writes to a multi-line tab separated values (tsv) file with the first 
    line having the feature names sorted alphabetically and separated by a tab '/t',
    and the rest of the lines containing the corresponding values for each participant
    separated by a tab '/t'
    For example:
    fixationrate    length    meanabspathangles
    0.0026852294    1529851    1.60354714212
    0.00456324344    453455    1.74324423
    
    Args:
        participants: a list of "Participant"s
        outfile: a string containing the name of the output file
        featurelist: if not None, a list of strings containing the name of the features to be
            returned
        
        aoifeaturelist: if not None, a list of features to be returned for each of the "AOI"s. 

        id_prefix: a boolean determining if the method should also export the participant id 
        
        require_valid: a boolean determining if only valid segments should be used when
        calculating the features. default = True
    """
    fnames, fvals = export_features_all(participants, featurelist =  featurelist, 
                                        aoifeaturelabels = aoifeaturelabels,
                                        aoifeaturelist = aoifeaturelist, id_prefix=id_prefix)
    
    with open(outfile, 'w') as f:
        f.write(string.join(fnames, '\t') + '\n')
        for l in fvals:
            f.write(string.join(map(str, l), '\t') + '\n')

def partition(segfile):
    """
    Placeholder for a method that generates the scenes based on some external log files and/or 'Event-Data.tsv" files
    """
    raise Exception("You should override the default Participant.partition method!")
    return

def test_validity():
    """
    Placeholder for a method that tests the quality of the eye data for each Participant
    """
    return

def read_events(evfile):
    """Returns a list of Event objects read from an 'Event-Data.tsv' file.

    Args:
        evfile: a string containing the name of the 'Event-Data.tsv' file exported from 
            Tobii software
    
    Returns:
        a list of Event objects
    """
    with open(evfile, 'r') as f:
        lines = f.readlines()

    return map(Event, lines[(params.EVENTSHEADERLINES+params.NUMBEROFEXTRAHEADERLINES):])



"""
UBC Eye Movement Data Analysis Toolkit

Module that contains all global parameters. Values of parameters are specific to the project.

Author: skardan

"""

EYELOGDATAFOLDER = "./sampledata"
# the folder that has the files exported from Tobii

EXTERNALLOGDATAFOLDER = "./sampledata/external logs"
# the folder that has the external log files

NUMBEROFEXTRAHEADERLINES = 8 
#number of extra lines at the beginning of the files exported from Tobii 
#this is specific to study and is based on the number of variables defined in Tobii
# studio for the experiment

FIXATIONHEADERLINES = 19
#number of lines at the beginning of the 'Fixation-Data' files exported from Tobii before 
#the actual data 

ALLDATAHEADERLINES = 26
#number of lines at the beginning of the 'All-Data' files exported from Tobii before 
#the actual data 

EVENTSHEADERLINES = 19
#number of lines at the beginning of the 'Event-Data' files exported from Tobii before 
#the actual data 

ACTIONHEADERLINES = 0
#number of lines at the beginning of the external log files before the actual data 

MEDIA_OFFSET = (0, 0)   
# the coordinates of the top left corner of the window
# showing the interface under study. (0,0) if the interface was
# in full screen (default value)

featurelist = ['numsegments','length','numfixations','fixationrate','meanabspathangles',
               'meanfixationduration','meanpathdistance','meanrelpathangles','stddevabspathangles',
               'stddevfixationduration','stddevpathdistance','stddevrelpathangles']#'numsamples','sumabspathangles','sumfixationduration','sumpathdistance','sumrelpathangles']
# list of non-AOI feature names 

aoigeneralfeat = ['fixationrate','numfixations','totaltimespent','proportionnum',
                  'proportiontime','longestfixation']#'timetofirstfixation','timetolastfixation',
#list of general AOI features

aoinames = ['Top','Bottom','Graph','Toolbar','Test']
#list of the AOI names

aoitransfrom = map(lambda x:'numtransfrom_'+x, aoinames) 
#['numtransfromBottom','numtransfromGraph','numtransfromToolbar','numtransfromTop']
#list of transition-based AOI features

#aoitransto = ['numtranstoBottom','numtranstoGraph','numtranstoToolbar','numtranstoTop']


aoiproportion = map(lambda x:'proptransfrom_'+x, aoinames) 
#['proptransfromBottom','proptransfromGraph','proptransfromToolbar','proptransfromTop']
#list of transition-based AOI features

##,'_proptransto_Bottom','_proptransto_Graph','_proptransto_Toolbar','_proptransto_Top']



#generating a list of all AOI-based features
aoifeaturelist =[]
for aoin in aoinames:
    aoifeaturelist.extend(map(lambda x:aoin+'_'+x, aoigeneralfeat))
    aoifeaturelist.extend(map(lambda x:aoin+'_'+x, aoitransfrom))
    #aoifeaturelist.extend(map(lambda x:aoin+x, aoitransto))
    aoifeaturelist.extend(map(lambda x:aoin+'_'+x, aoiproportion))
    

# Validity Threshold for segments
VALID_PROP_THRESH = 0.85
#the minimum proportion of valid samples for a Segment or Scene to be considered valid

VALID_TIME_THRESH = 3000
#the maximum gap size (ms) allowable in samples for a Segment or Scene to be considered valid

MAX_SEG_TIMEGAP = 300 #maximum gap size (ms) allowable in a segment with auto-partition option
"""
Auto-partition option is when EMDAT automatically splits the "Segment"s which have low sample 
quality, into two new sub "Segment"s discarding the largest gap of invalid samples for 
a "Segment". EMDAT will continue to perform the splitting on the Segments until there is not
any gap larger than MAX_SEG_TIMEGAP left in the data.
"""

VALIDITY_METHOD = 3 #1: porportion; 2:time gap; 3: porportion with (valid + restored) samples
"""
Methods 1 and 3 use VALID_PROP_THRESH and method 2 uses VALID_TIME_THRESH as validity threshold
 
Restored samples are the samples which are not valid but they are part of a Fixation.
The idea is that if the user was looking at a certain point and then we loose the eye data for 
a short period of time and afterwards the user is looking at the same point we can assume that user
was looking at that same point during that period. 
"""

MINSEGSIZE = 300 #minimum segment size in ms that is considered meaningful for this experiment

INCLUDE_HALF_FIXATIONS = False
#A boolean value determining if a Fixation extends between two consecutive Segments, should it be 
#included in those Segments or not

DEBUG = False
#DEBUG = True
#Enable/disable verbose mode which prints out more information for debugging purposes



#Predefined lists of features that are not calculated using time
NONTEMP_FEATURES_SEG = ['meanfixationduration', 'stddevfixationduration',
'fixationrate', 'meanpathdistance', 'stddevpathdistance', 'meanabspathangles', 'stddevabspathangles',
'meanrelpathangles', 'stddevabspathangles'] 

NONTEMP_FEATURES_AOI = ['longestfixation', 'proportionnum', 'proportiontime',
'proptransto', 'proptransfrom']



"""
UBC Eye Movement Data Analysis Toolkit
Created on 2011-08-26

@author: skardan


"""
import skgaze.EMDAT.params as params
import skgaze.EMDAT.geometry as geometry
from skgaze.EMDAT.AOI import *
from warnings import warn



class Segment():
    """A Segment is a class that represents the smallest unit of aggregated eye data samples with a conceptual meaning.

    A segment is the smallest unit of aggregated eye data samples that has conceptual meaning. This class is the equivalent
    of segments as defined in Tobii studio. 
    
    Attributes:
        segid: A string containing the id of the Segment.
        alldata: A list of "Datapoint"s for this Segment
        features: A dict with feature names as its keys and feature values as its values 
        completion_time: An integer indicating total duration of the Segment in milliseconds
            minimum is 16 ms (length of one sample with 60Hz sampling rate (ms))
        start: An integer indicating the Segment's start time in milliseconds
        end: An integer indicating the Segment's end time in milliseconds
        sample_start_ind: An integer indicating the index of the first Datapoint for this Segment in the Participant's list of all "Datapoint"s (all_data) 
        sample_end_ind: An integer indicating the index of the last Datapoint for this Segment in the Participant's list of all "Datapoint"s (all_data)
        fixation_start_ind: An integer indicating the index of the first Fixation for this Segment in the Participant's list of all "Fixation"s (fixation_data)
        fixation_end_ind: An integer indicating the index of the first Fixation for this Segment in the Participant's list of all "Fixation"s (fixation_data)
        numfixations: An integer indicating the number of "Fixation"s in this Segment
        time_gaps: a list of tuples of the form (start, end) indicating the start and end of the gaps of invalid samples in the Segement's samples
        largest_data_gap: An integer indicating the length of largest invalid gap for this Segment in milliseconds
        proportion_valid: A float indicating the proportion of valid samples over all the samples in this Segment
        proportion_valid_fix: A float indicating the proportion of (valid + restored) samples over all the samples in this Segment 
        validity1: a boolean indicating whether this Segment is valid using proportion of valid samples threshold 
        validity2: a boolean indicating whether this Segment is valid using largest acceptable gap threshold
        validity3: a boolean indicating whether this Segment is valid using proportion of (valid + restored) samples threshold
        is_valid: a boolean indicating whether this Segment is considered valid by the validity method indicated by params.VALIDITY_METHOD
        length: An integer indicating total duration of the Segment in milliseconds
        numsamples: An integer indicating total number of samples in the Segment 
        fixation_data: A list of "Fixation"s for this Segment
        fixation_start: timestamp of the first entry from list of "Fixation"s for this Segment
        fixation_end: timestamp of the last entry from list of "Fixation"s for this Segment
        aoi_data: A list of AOI_Stat objects for relevant "AOI"s for this Segment
        has_aois: A boolean indicating if this Segment has AOI features calculated for it
        
    """
    def __init__(self, segid, all_data, fixation_data, aois = None, prune_length = None):
        """
        Args:
            segid: A string containing the id of the Segment.
            
            all_data: a list of "Datapoint"s which make up this Segment.
            
            fixation_data: a list of "Fixation"s which make up this Segment.
            
            aois: a list of "AOI"s relevant to this Segment.
         
            prune_length: If not None, an integer that specifies the time interval (in ms) from the beginning of each segment in which
                samples are considered in calculations.  This can be used if, for example, you only wish to consider data in the first 
                1000 ms of each segment. In this case (prune_length = 1000), all data beyond the first 1000ms of the start of the segments
                will be disregarded.
                
        Yields:
            a Segment object
        """
        self.segid = segid
        #self.alldata = all_data
        self.features = {}
        
        self.completion_time = int(all_data[-1].timestamp) - int(all_data[0].timestamp)
        if self.completion_time == 0:
            raise Exception("Zero length segment")
#            self.completion_time = 16 #length of one sample with 60Hz sampling rate (ms)
#        for d in all_data:
#            d.set_segid(segid)
        self.features['completion_time'] = self.completion_time
        self.start = all_data[0].timestamp
        self.numfixations = len(fixation_data)
        self.time_gaps = []
        self.largest_data_gap = self.calc_largest_validity_gap(all_data)
        self.proportion_valid = self.calc_validity_proportion(all_data)
        self.proportion_valid_fix = self.calc_validity_fixation(all_data)
        self.validity1 = self.calc_validity1()
        self.validity2 = self.calc_validity2()
        self.validity3 = self.calc_validity3()
        self.is_valid = self.get_validity()
        if prune_length:
            all_data = filter(lambda x: x.timestamp <= self.start +
            prune_length, all_data)
            fixation_data = filter(lambda x: x.timestamp <= self.start +
            prune_length, fixation_data)
        self.end = all_data[-1].timestamp
        self.length = int(self.end) - int(self.start)
        self.features['length'] = int(self.end)- int(self.start)
        self.numsamples = self.calc_num_samples(all_data)
        self.features['numsamples'] = self.numsamples
        self.numfixations = len(fixation_data)
#        for f in fixation_data:
#            f.set_segid(segid)
        self.features['numfixations'] = self.numfixations
        self.features['fixationrate'] = float(self.numfixations) / self.length
        if self.numfixations > 0:
            self.fixation_start = fixation_data[0].timestamp
            self.fixation_end = fixation_data[-1].timestamp
            self.features['meanfixationduration'] = mean(list(map(lambda x: float(x.fixationduration), fixation_data)))
            self.features['stddevfixationduration'] = stddev(list(map(lambda x: float(x.fixationduration), fixation_data)))
            self.features['sumfixationduration'] = sum(map(lambda x: x.fixationduration, fixation_data))
            self.features['fixationrate'] = float(self.numfixations)/self.length
            distances = self.calc_distances(fixation_data)
            abs_angles = self.calc_abs_angles(fixation_data)
            rel_angles = self.calc_rel_angles(fixation_data)
        else:
            self.fixation_start = -1
            self.fixation_end = -1            
            self.features['meanfixationduration'] = 0
            self.features['stddevfixationduration'] = 0
            self.features['sumfixationduration'] = 0
            self.features['fixationrate'] = 0
            distances = []
        if len(distances) > 0:
            self.features['meanpathdistance'] = mean(distances)
            self.features['sumpathdistance'] = sum(distances)
            self.features['stddevpathdistance'] = stddev(distances)
            self.features['eyemovementvelocity'] = self.features['sumpathdistance']/self.length
            self.features['sumabspathangles'] = sum(abs_angles)
            self.features['meanabspathangles'] = mean(abs_angles)
            self.features['stddevabspathangles'] = stddev(abs_angles)
            self.features['sumrelpathangles'] = sum(rel_angles)
            self.features['meanrelpathangles'] = mean(rel_angles)
            self.features['stddevrelpathangles'] = stddev(rel_angles)
        else:
            self.features['meanpathdistance'] = 0
            self.features['sumpathdistance'] = 0
            self.features['stddevpathdistance'] = 0
            self.features['sumabspathangles'] = 0
            self.features['meanabspathangles']= 0
            self.features['stddevabspathangles']= 0
            self.features['sumrelpathangles'] = 0
            self.features['meanrelpathangles']= 0
            self.features['stddevrelpathangles'] = 0
        self.has_aois = False
        if aois:
            self.set_aois(aois,fixation_data)
    def set_indices(self,sample_st,sample_end,fix_st,fix_end):
        """Sets the index features
        
        Args:
            sample_st: An integer indicating the index of the first Datapoint for this Segment in the Participant's list of all "Datapoint"s (all_data) 
            sample_end: An integer indicating the index of the last Datapoint for this Segment in the Participant's list of all "Datapoint"s (all_data)
            fix_st: An integer indicating the index of the first Fixation for this Segment in the Participant's list of all "Fixation"s (fixation_data)
            fix_st: An integer indicating the index of the first Fixation for this Segment in the Participant's list of all "Fixation"s (fixation_data)        
        """
        self.sample_start_ind = sample_st
        self.sample_end_ind = sample_end
        self.fixation_start_ind = fix_st
        self.fixation_end_ind = fix_end

    def get_indices(self):
        """Returns the index features
        
        Returns:
            An integer indicating the index of the first Datapoint for this Segment in the Participant's list of all "Datapoint"s (all_data) 
            An integer indicating the index of the last Datapoint for this Segment in the Participant's list of all "Datapoint"s (all_data)
            An integer indicating the index of the first Fixation for this Segment in the Participant's list of all "Fixation"s (fixation_data)
            An integer indicating the index of the first Fixation for this Segment in the Participant's list of all "Fixation"s (fixation_data)   
            
        Raises:
            Exception: An exception is thrown if the values are read before initialization
        """
        if self.sample_start_ind != None:
            return self.sample_start_ind, self.sample_end_ind, self.fixation_start_ind, self.fixation_end_ind 
        raise Exception ('The indices values are accessed before setting the initial value in segement:'+self.segid+'!')

    def set_aois(self, aois, fixation_data):
        """Sets the relevant "AOI"s for this Segment
        
        Args:
            fixation_data: The list of "Fixation"s which make up this Segment
            aois: a list of "AOI"s relevant to this Segment
        """
        
        if len(aois) == 0:
            warn("no AOIs passed to segment:"+self.segid)
        active_aois=[]
        for aoi in aois:
            #print "checking:",aoi.aid
            if aoi.is_active(self.fixation_start, self.fixation_end):
                active_aois.append(aoi)
        if not(active_aois):
            msg = "no active AOIs passed to segment:%s start:%d end:%d" %(self.segid,self.start,self.end)
            warn(msg)
        self.aoi_data = {}
        for aoi in active_aois:
            aoistat = AOI_Stat(aoi, fixation_data, self.start, self.end, active_aois)
            self.aoi_data[aoi.aid] = aoistat
            self.has_aois = True

    def calc_validity_proportion(self, all_data):
        """Calculates the proportion of "Datapoint"s which are valid.
        
        Args:
            all_data: The list of "Datapoint"s which make up this Segement
            
        Returns:
            A float indicating the proportion of valid samples over all the samples in this Segment
        """
        num_valid = float(0)
        num = 0

        for d in all_data:
            #if d.stimuliname == 'ScreenRec':
            if d.stimuliname != '':
                num += 1
                if d.is_valid:
                    num_valid += 1
#            else:
#                print "###",d.event, d.data1
        if num==0:
            return 0.0
        else:
            return num_valid / num

    def calc_largest_validity_gap(self, all_data):
        """Calculates the largest gap of invalid samples in the "Datapoint"s for this Segment.
        
        Args:
            all_data: The list of "Datapoint"s which make up this Segement
            
        Returns:
            An integer indicating the length of largest invalid gap for this Segment in milliseconds  
        """
        if self.numfixations == 0:
            return all_data[-1].timestamp - all_data[0].timestamp
        self.time_gaps = []
        max = 0
        dindex = 0
        datalen = len(all_data)
        while dindex < datalen:
            d = all_data[dindex]
            while d.is_valid and (dindex < datalen-1):
                dindex += 1
                d = all_data[dindex]
            if not(d.is_valid):
                gap_start = d.timestamp
                while not(d.is_valid) and (dindex < datalen-1):
                    dindex += 1
                    d = all_data[dindex]
                if int(d.timestamp) - int(gap_start) > max:
                    max = int(d.timestamp) - int(gap_start)
                if int(d.timestamp) - int(gap_start) > params.MAX_SEG_TIMEGAP:
                    self.time_gaps.append((gap_start,d.timestamp))
            dindex += 1

        return max

    def getgaps(self):
        """Returns the list of invalid gaps for this Segment
        
        Args:
            a list of invalid gaps for this Segment
        """
        return self.time_gaps

    def calc_validity_fixation(self, all_data):
        """Calculates the proportion of (valid + restored) "Datapoint"s over all "Datapoint"s of the Segment.
        
        Restored samples are the samples which are not valid but they are part of a Fixation.
        The idea is that if the user was looking at a certain point and then we loose the eye data for 
        a short period of time and afterwards the user is looking at the same point we can assume that user
        was looking at that same point during that period. 
        
        Args:
            all_data: The list of "Datapoint"s which make up this Segement
            
        Returns:
            A float indicating the proportion of (valid + restored) samples over all the samples in this Segment
        """
        if self.numfixations == 0:
            return 0.0
        num_valid = float(0)
        num = 0

        for d in all_data:
            #if d.stimuliname == 'ScreenRec':
            if d.stimuliname != '':
                num += 1
                if d.fixationindex!=None:
                    num_valid += 1.0
#            else:
#                print "###",d.event, d.data1
        if num==0:
            return 0.0
        else:
            return num_valid / num
        
    def calc_validity1(self, threshold = params.VALID_PROP_THRESH):
        """Returns a boolean indicating whether this Segment is valid using proportion of valid samples threshold
        
        Args:
            threshold: the minimum proportion of valid samples for a Segment or Scene to be 
                considered valid. By default set to value VALID_PROP_THRESH from module params.py 
        """
        return self.proportion_valid > threshold

    def calc_validity2(self, threshold = params.VALID_TIME_THRESH):
        """Returns a boolean indicating whether this Segment is valid using largest acceptable gap threshold
        """
        return self.largest_data_gap <= threshold

       
    def calc_validity3(self, threshold = params.VALID_PROP_THRESH):
        """Returns a boolean indicating whether this Segment is valid using proportion of (valid + restored) samples threshold
        """
        return self.proportion_valid_fix > threshold
    
    def get_validity(self):
        """Determines if this Segment is valid with the given validity method set in params.VALIDITY_METHOD
        
        Returns:
            A boolean indicating whether this Segment is valid
        """
        if params.VALIDITY_METHOD == 1:
            return self.validity1
        elif params.VALIDITY_METHOD == 2:
            return self.validity2
        elif params.VALIDITY_METHOD == 3:
            return self.validity3
    
    def calc_distances(self, fixdata):
        """returns the Euclidean distances between a sequence of "Fixation"s
    
        Args:
            fixdata: a list of "Fixation"s
        """
        distances = []
        lastx = fixdata[0].mappedfixationpointx
        lasty = fixdata[0].mappedfixationpointy

        for i in range(1, len(fixdata)):
            x = fixdata[i].mappedfixationpointx
            y = fixdata[i].mappedfixationpointy
            dist = math.sqrt((x-lastx)**2 + (y-lasty)**2)
            distances.append(dist)
            lastx = x
            lasty = y

        return distances

    def calc_abs_angles(self, fixdata):
        """returns the absolute angles between a sequence of "Fixation"s that build a scan path.
        
        Abosolute angle for each saccade is the angle between that saccade and the horizental axis
    
        Args:
            fixdata: a list of "Fixation"s
            
        Returns:
            a list of absolute angles for the saccades formed by the given sequence of "Fixation"s in Radiant
        """
        abs_angles = []
        lastx = fixdata[0].mappedfixationpointx
        lasty = fixdata[0].mappedfixationpointy

        for i in range(1,len(fixdata)):
            x = fixdata[i].mappedfixationpointx
            y = fixdata[i].mappedfixationpointy
            (dist, theta) = geometry.vector_difference((lastx,lasty), (x, y))
            abs_angles.append(abs(theta))
            lastx=x
            lasty=y

        return abs_angles

    def calc_rel_angles(self, fixdata):
        """returns the relative angles between a sequence of "Fixation"s that build a scan path in Radiant
        
        Relative angle for each saccade is the angle between that saccade and the previous saccade.
    
        Args:
            fixdata: a list of "Fixation"s
            
        Returns:
            a list of relative angles for the saccades formed by the given sequence of "Fixation"s in Radiant
        """
        rel_angles = []
        lastx = fixdata[0].mappedfixationpointx
        lasty = fixdata[0].mappedfixationpointy

        for i in range(1, len(fixdata)-1):
            x = fixdata[i].mappedfixationpointx
            y = fixdata[i].mappedfixationpointy
            nextx = fixdata[i+1].mappedfixationpointx
            nexty = fixdata[i+1].mappedfixationpointy
            (dist, theta) = geometry.vector_difference((x,y), (lastx, lasty))
            (dist, nextheta) = geometry.vector_difference((x,y), (nextx, nexty))
            theta = abs(theta-nextheta)
            rel_angles.append(theta)
            lastx=x
            lasty=y

        return rel_angles

    def calc_num_samples(self, all_data):
        """Returns the number of samples in the Segment
        
        Args:
            all_data: a list of "Datapoint"s which make up this Segment.
            
        Returns:
            An integer determining the number of samples in the Segment
        
        """
        num = 0
        for d in all_data:
            if d.stimuliname != '':
                num += 1
        return num
    
    def getid(self):
        """Returns the segid for this Segment
        
        Returns: a string conataining the segid for this Segment
        """
        return self.segid
    
    def get_features(self, featurelist = None, aoifeaturelist = None, aoifeaturelabels = None):
        """Returns feature names and their values for this Segment
        
        Args:
            featurelist: if not None, a list containing the name of features to be returned. If this is None all features will be returned
            aoifeaturelist: if not None, a list of features to be returned for each of the "AOI"s relevant to this Segment. 
            aoifeaturelabels: if not None, a list of AOI related features to be returned.
            *Note: while aoifeaturelist is a subset of features that will be returned for all relevant "AOI"s, aoifeaturelabels contains 
            the exact AOI feature name, i.e., a feature of the form: [AOI name]_[feature name]
            For example if an AOI called 'graph' is releveant to this Segment, aoifeaturelabels may contain 'graph_fixationrate'  
            
        Returns:
            featnames: a list of feature names sorted alphabetically
            featvals: a corrsponding list of feature values
            e.g.
            featnames = ['fixationrate', 'length', 'meanabspathangles']
            featvals  = [0.00268522882294', '1529851', '1.60354714212']
        
        """ 
        if featurelist == []:
            featnames = []
        elif not featurelist:       #include all features
            featnames = self.features.keys()
        else:                       #a list of features was given
            featnames = []
            for name in featurelist:
                if name in self.features.keys():
                    featnames.append(name)
                else:
                    raise Exception('Segement %s has no such feature: %s'%(self.getid(),name))

        featnames = sorted(featnames)

        featvals = list(map(lambda x: self.features[x], featnames))

        if self.has_aois:
            for aid, aoi in self.aoi_data.items():
                if aoifeaturelabels:    #an exact list of aoifeatures was given 
                    anames, avals = aoi.get_features()
                    anames = map(lambda x: '%s_%s'%(aid, x), anames)
                    featval = zip(anames,avals)
                    newfeatval = filter(lambda x: x[0] in aoifeaturelabels,featval)
                    anames = []
                    avals = []
                    for fn,fv in newfeatval:
                        anames.append(fn)
                        avals.append(fv)
                    if featnames:                    
                        featnames += anames
                        featvals += avals                   
                else:                   #a list of features for each AIO was given
                    anames, avals = aoi.get_features(aoifeaturelist)
                    anames = list(map(lambda x: '%s_%s'%(aid, x), anames))
                    featnames += anames
                    featvals += avals

        return featnames, featvals
    
    def print_(self):
        """Ourputs all feature names and their values for this Segment on the console        
        """ 
        print("ID", self.getid())
        print("start",self.start)
        print("end",self.end)
        print("is_valid",self.is_valid)
        
#        featurelist =["completion_time","numfixations","length","numsamples"]
#        if self.features['numfixations'] > 0:
#            featurelist.extend(["meanfixationduration","stddevfixationduration","sumfixationduration","fixationrate"])
#        
#        featurelist.extend(["meanpathdistance","sumpathdistance","stddevpathdistance","sumabspathangles","meanabspathangles","stddevabspathangles","sumrelpathangles","meanrelpathangles","stddevrelpathangles"])

        fn,fv = self.get_features()
        for i in range(len(fn)):
            print(fn[i],':',fv[i])
        print()
    

        



from skgaze.RQA.RQA_funcs import RQA as _RQA

class RQA:

    def __init__(self, dataset):
        self.dataset = dataset
        
    
    def CalculateRecurrenceFunction(self,timeDelayValue, numTimeDelaySample, phaseSpaceClusteringThreshold, pid=False):

        result = {}
        if pid:
            for participant in self.dataset.participants:
                if participant.id == pid:
                    return _RQA(participant.fixations).CalculateRecurrenceFunction(timeDelayValue, numTimeDelaySample, phaseSpaceClusteringThreshold)
                print("couldn't find participant with this id")
                return None
        
        for participant in self.dataset.participants:
            result[participant.id] = _RQA(participant.fixations).CalculateRecurrenceFunction(timeDelayValue, numTimeDelaySample, phaseSpaceClusteringThreshold)
        return result

    def CalculateReoccurrenceFunction(self, fixationSpaceDistanceRadius, pid=False):

        result = {}
        if pid:
            for participant in self.dataset.participants:
                if participant.id == pid:
                    return _RQA(participant.fixations).CalculateReoccurrenceFunction(fixationSpaceDistanceRadius)
                print("couldn't find participant with this id")
                return None
        
        for participant in self.dataset.participants:
            result[participant.id] = _RQA(participant.fixations).CalculateReoccurrenceFunction(fixationSpaceDistanceRadius)
        return result



from skgaze.similarity_algorithms import similarity	

class STA:
    
    def __init__(self, dataset):
        self.raw_sequences = dataset.create_raw_sequences()
        self.aoi_data = self.create_aoi_array(dataset.aoi_array)

    def create_aoi_array(self, aoi_array):
    	sta_aoi_array = []
    	for aoi in aoi_array:
        	sta_aoi_array.append([aoi.type_of_element,float(aoi.x),float(aoi.width),float(aoi.y),float(aoi.height),aoi.aoi_char])
    	return sta_aoi_array

    def getNumberedSequence(self, Sequence, aoi_data):
	    numberedSequence = []
	    numberedSequence.append([Sequence[0][0], 1, Sequence[0][1]])
	
	    for y in range(1, len(Sequence)):
	        if Sequence[y][0] == Sequence[y - 1][0]:
	            numberedSequence.append([Sequence[y][0], numberedSequence[len(numberedSequence) - 1][1], Sequence[y][1]])
	        else:
	            numberedSequence.append([Sequence[y][0], self.getSequenceNumber(Sequence[0:y], Sequence[y][0]), Sequence[y][1]])
	
	    AoIList = self.getExistingAoIListForSequence(numberedSequence)
	    AoINames = aoi_data
	    AoINames = [w[5] for w in AoINames]
	    newSequence = []
	
	    myList = []
	    myDictionary = {}
	    replacementList = []
	
	    for x in range(0, len(AoIList)):
	        totalDuration = 0
	        for y in range(0, len(numberedSequence)):
	            if numberedSequence[y][0:2] == AoIList[x]:
	                totalDuration = totalDuration + int(numberedSequence[y][2])
	        myList.append([AoIList[x], totalDuration])
	
	    for x in range(0, len(AoINames)):
	        myAoIList = [w for w in myList if w[0][0] == AoINames[x]]
	        myAoIList.sort(key=lambda x: x[1])
	        myAoIList.reverse()
	        if len(myAoIList) > 0:
	            myDictionary[AoINames[x]] = myAoIList
	
	    for AoI in AoIList:
	        index = [w[0] for w in myDictionary[AoI[0]]].index(AoI)
	        replacementList.append([AoI, [AoI[0], (index + 1)]])
	
	    for x in range(0, len(numberedSequence)):
	        myReplacementList = [w[0] for w in replacementList]
	        index = myReplacementList.index(numberedSequence[x][0:2])
	        newSequence.append([replacementList[index][1][0]] + [replacementList[index][1][1]] + [numberedSequence[x][2]])
	
	    return newSequence
	
	
    def getSequenceNumber(self,Sequence, Item):
	    abstractedSequence = self.getAbstractedSequence(Sequence)
	    return abstractedSequence.count(Item) + 1
	
	
    def getAbstractedSequence(self,Sequence):
	    myAbstractedSequence = [Sequence[0]]
	    for y in range(1, len(Sequence)):
	        if myAbstractedSequence[len(myAbstractedSequence) - 1] != Sequence[y][0]:
	            myAbstractedSequence.append([Sequence[y][0], Sequence[y][1]])
	    return myAbstractedSequence
	
	
    def getExistingAoIListForSequence(self,Sequence):
	    AoIlist = []
	    for x in range(0, len(Sequence)):
	        try:
	            AoIlist.index(Sequence[x][0:2])
	        except:
	            AoIlist.append(Sequence[x][0:2])
	    return AoIlist
	
	
    def calculateImportanceThreshold(self,mySequences):
	    myAoICounter = self.getNumberDurationOfAoIs(mySequences)
	    commonAoIs = []
	    for myAoIdetail in myAoICounter:
	        if myAoIdetail[3] == True:
	            commonAoIs.append(myAoIdetail)
	
	    if len(commonAoIs) == 0:
	        print("No shared instances!")
	        exit(1)
	
	    minValueCounter = commonAoIs[0][1]
	    for AoIdetails in commonAoIs:
	        if minValueCounter > AoIdetails[1]:
	            minValueCounter = AoIdetails[1]
	
	    minValueDuration = commonAoIs[0][2]
	    for AoIdetails in commonAoIs:
	        if minValueDuration > AoIdetails[2]:
	            minValueDuration = AoIdetails[2]
	
	    return [minValueCounter, minValueDuration]
	
	
    def getNumberDurationOfAoIs(self,Sequences):
	    AoIs = self.getExistingAoIList(Sequences)
	    AoIcount = []
	    for x in range(0, len(AoIs)):
	        counter = 0
	        duration = 0
	        flagCounter = 0
	        keys = list(Sequences.keys())
	        for y in range(0, len(keys)):
	            if [s[0:2] for s in Sequences[keys[y]]].count(AoIs[x]) > 0:
	                counter = counter + [s[0:2] for s in Sequences[keys[y]]].count(AoIs[x])
	                duration = duration + sum([int(w[2]) for w in Sequences[keys[y]] if w[0:2] == AoIs[x]])
	                flagCounter = flagCounter + 1
	
	        if flagCounter > len(keys) / 2:
	            AoIcount.append([AoIs[x], counter, duration, True])
	        else:
	            AoIcount.append([AoIs[x], counter, duration, False])
	    return AoIcount
	
	
    def updateAoIsFlag(self, AoIs, threshold):
	    for AoI in AoIs:
	        if AoI[1] >= threshold[0] and AoI[2] >= threshold[1]:
	            AoI[3] = True
	    return AoIs
	
	
    def removeInsignificantAoIs(self, Sequences, AoIList):
	    significantAoIs = []
	    for AoI in AoIList:
	        if AoI[3] == True:
	            significantAoIs.append(AoI[0])
	
	    keys = list(Sequences.keys())
	    for y in range(0, len(keys)):
	        temp = []
	        for k in range(0, len(Sequences[keys[y]])):
	            try:
	                significantAoIs.index(Sequences[keys[y]][k][0:2])
	                temp.append(Sequences[keys[y]][k])
	            except:
	                continue
	        # Fix for scanpaths consisting only from insignificant aois - remove such scanpaths
	        if len(temp) > 0:
	            Sequences[keys[y]] = temp
	        else:
	            Sequences.pop(keys[y], None)
	    return Sequences
	
	
    def getExistingAoIList(self, Sequences):
	    AoIlist = []
	    keys = list(Sequences.keys())
	    for y in range(0, len(keys)):
	        for x in range(0, len(Sequences[keys[y]])):
	            try:
	                AoIlist.index(Sequences[keys[y]][x][0:2])
	            except:
	                AoIlist.append(Sequences[keys[y]][x][0:2])
	    return AoIlist
	
	
    def calculateNumberDurationOfFixationsAndNSV(self, Sequences):
	    keys = list(Sequences.keys())
	    for x in range(0, len(keys)):
	        myAbstractedSequence = []
	        myAbstractedSequence = [Sequences[keys[x]][0][0:2] + [1] + [int(Sequences[keys[x]][0][2])]]
	        for y in range(1, len(Sequences[keys[x]])):
	            if myAbstractedSequence[len(myAbstractedSequence) - 1][0:2] != Sequences[keys[x]][y][0:2]:
	                myAbstractedSequence.append(Sequences[keys[x]][y][0:2] + [1] + [int(Sequences[keys[x]][y][2])])
	            else:
	                myAbstractedSequence[len(myAbstractedSequence) - 1][2] = \
	                myAbstractedSequence[len(myAbstractedSequence) - 1][2] + 1
	                myAbstractedSequence[len(myAbstractedSequence) - 1][3] = \
	                myAbstractedSequence[len(myAbstractedSequence) - 1][3] + int(Sequences[keys[x]][y][2])
	
	        Sequences[keys[x]] = myAbstractedSequence
	
	    keys = list(Sequences.keys())
	
	    for x in range(0, len(keys)):
	        for y in range(0, len(Sequences[keys[x]])):
	            if len(Sequences[keys[x]]) < 2:
	                value = 0
	            else:
	                value = 0.9 / (len(Sequences[keys[x]]) - 1)
	            NSV = 1 - round(y, 2) * value
	            Sequences[keys[x]][y] = Sequences[keys[x]][y] + [NSV]
	    return Sequences
	
	
    def calculateTotalNumberDurationofFixationsandNSV(self, AoIList, Sequences):
	    for x in range(0, len(AoIList)):
	        duration = 0
	        counter = 0
	        totalNSV = 0
	
	        flag = 0
	        keys = list(Sequences.keys())
	        for y in range(0, len(keys)):
	            for k in range(0, len(Sequences[keys[y]])):
	                if Sequences[keys[y]][k][0:2] == AoIList[x]:
	                    counter += Sequences[keys[y]][k][2]
	                    duration += Sequences[keys[y]][k][3]
	                    totalNSV += Sequences[keys[y]][k][4]
	                    flag += 1
	        if flag == len(Sequences):
	            AoIList[x] = AoIList[x] + [counter] + [duration] + [totalNSV] + [True]
	        else:
	            AoIList[x] = AoIList[x] + [counter] + [duration] + [totalNSV] + [False]
	
	    return AoIList
	
	
    def getValueableAoIs(self, AoIList):
	    commonAoIs = []
	    valuableAoIs = []
	    for myAoIdetail in AoIList:
	        if myAoIdetail[5] == True:
	            commonAoIs.append(myAoIdetail)
	
	    minValue = commonAoIs[0][4] if len(commonAoIs) > 0 else 0
	    for AoIdetails in commonAoIs:
	        if minValue > AoIdetails[4]:
	            minValue = AoIdetails[4]
	
	    for myAoIdetail in AoIList:
	        if myAoIdetail[4] >= minValue:
	            valuableAoIs.append(myAoIdetail)
	
	    return valuableAoIs
	
    def format_sequences(self, raw_sequences):
	    """
	    {'01': [[A, 150], [B, 250]], '02': ...} gets transformed into:
	    [{'identifier': '01', 'fixations': [[A, 150], [B, 250]]}, {'identifier': '02' ... }]
	    """
	    formatted_sequences = []
	    keys = list(raw_sequences.keys())
	    for it in range(0, len(raw_sequences)):
	        act_rec = {
	            'identifier': keys[it],
	            'fixations': raw_sequences[keys[it]]
	        }
	        formatted_sequences.append(act_rec)
	
	    return formatted_sequences
	# STA Algorithm
    def run_sta(self):
	    """
	    Args:
	        raw_sequences: a Python dict of lists - {'ID1': [['F', '383'], ['G', '150']], 'ID2': .. }
	        aoi_data: for additional scanpath processing - [['fullAoiName', xFrom, width, yFrom, height, 'shortName'] .. ]
	    Returns:
	        identifier: for client-side purposes
	        fixations: a list of lists representing the common scanpath - [['A', 150], ['B', 500] .. ]
	        similarity: a dict containing similarity of individual scanpaths to the common one - {'ID1': 66.66, 'ID2': ... }
	    """
	
	    # Preliminary Stage is already complete - sequences are passed in via arguments
	
	    # First-Pass
	    mySequences_num = {}
	    keys = list(self.raw_sequences.keys())
	    for y in range(0, len(keys)):
	        mySequences_num[keys[y]] = self.getNumberedSequence(self.raw_sequences[keys[y]], self.aoi_data)
	
	    myImportanceThreshold = self.calculateImportanceThreshold(mySequences_num)
	    myImportantAoIs = self.updateAoIsFlag(self.getNumberDurationOfAoIs(mySequences_num), myImportanceThreshold)
	    myNewSequences = self.removeInsignificantAoIs(mySequences_num, myImportantAoIs)
	    # Second-Pass
	    myNewAoIList = self.getExistingAoIList(myNewSequences)
	    myNewAoIList = self.calculateTotalNumberDurationofFixationsandNSV(
	        myNewAoIList,
	        self.calculateNumberDurationOfFixationsAndNSV(myNewSequences)
	    )
	    myFinalList = self.getValueableAoIs(myNewAoIList)
	
	    myFinalList.sort(key=lambda x: (x[4], x[3], x[2]))
	    myFinalList.reverse()
	
	    commonSequence = []
	
	    for y in range(0, len(myFinalList)):
	        commonSequence.append([myFinalList[y][0], int(myFinalList[y][3] / myFinalList[y][2])])
	
	    formatted_sequences = self.format_sequences(self.raw_sequences)
	
	    # Store scanpaths as an array of string-converted original scanpaths
	    scanpath_strs = similarity.convert_to_str_array(formatted_sequences)
	
	    common_scanpath = self.getAbstractedSequence(commonSequence)
	    common_scanpath_str = ''
	
	    # For determining get_edit_distance distance we need a pure string version of the common scanpath ('ABC')
	    for fixation in common_scanpath:
	        common_scanpath_str += fixation[0]
	
	    res_data = {
	        'identifier': 'STA',
	        'fixations': common_scanpath,
	        'similarity': similarity.calc_similarity_to_common(scanpath_strs, common_scanpath_str)
	    }
	
	    return res_data
	


import copy
import math
from operator import itemgetter
from skgaze.similarity_algorithms import similarity
	
class Dotplots:
    
    def __init__(self, dataset):
        self.dotplot_max_aois = 100
        self.dotplot_error_rate_area = 0
        self.raw_sequences = dataset.create_raw_sequences()
	
	
	# CODE ORIGIN: https://github.com/Groosling/DP_features_and_scanpaths
    def createSequencesBasedOnVisualElements(self, my_dataset):
	    Sequences = {}
	    Participants = my_dataset.scanpath_data_raw
	    myAoIs = my_dataset.aoi_data
	    keys = list(Participants.keys())
	    
	
	    for y in range(0, len(keys)):
	        sequence = ""
	        counter = 0
	        for z in range(0, len(Participants[keys[y]])):
	            if counter == self.dotplot_max_aois:
	                break
	            tempAoI = ""
	            tempDuration = 0
	
	            for k in range(0, len(myAoIs)):
	                if float(Participants[keys[y]][z][3]) >= (float(myAoIs[k][1]) - self.dotplot_error_rate_area) and float(
	                        Participants[keys[y]][z][3]) < (
	                ((float(myAoIs[k][1]) - self.dotplot_error_rate_area) + (float(myAoIs[k][2]) + 2 * self.dotplot_error_rate_area))) and float(
	                        Participants[keys[y]][z][4]) >= (float(myAoIs[k][3]) - self.dotplot_error_rate_area) and float(
	                        Participants[keys[y]][z][4]) < (
	                ((float(myAoIs[k][3]) - self.dotplot_error_rate_area) + (float(myAoIs[k][4]) + 2 * self.dotplot_error_rate_area))):
	                    tempAoI = tempAoI + myAoIs[k][5]
	                    tempDuration = int(Participants[keys[y]][z][2])
	
	
	            # my solution compare sum of distances to four corners
	            if len(tempAoI) > 1:
	                tempAoI = self.getCloserAOI(Participants[keys[y]][z],myAoIs, tempAoI)
	
	            if len(tempAoI) != 0:
	                counter = counter + 1
	                sequence = sequence + tempAoI + "-" + str(tempDuration) + "."
	                if counter == self.dotplot_max_aois:
	                    break
	
	        Sequences[keys[y]] = sequence
	    return Sequences
	
	
    def getCloserAOI(self,Participants_pos, myAoIs, tempAoI):
	    sums_of_distances = {}
	    for m in range(0, len(tempAoI)):
	        for n in range(0, len(myAoIs)):
	            if tempAoI[m] == myAoIs[n][5]:
	                temp_distance = []
	                # sum distance of all 4 corners
	                # up, left
	                temp_distance.append(math.sqrt(pow(float(Participants_pos[3]) - float(myAoIs[n][1]), 2) +
	                                               pow(float(Participants_pos[4]) - float(myAoIs[n][3]), 2)))
	                # up right
	                temp_distance.append(math.sqrt(pow(float(Participants_pos[3]) - (float(myAoIs[n][1]) + float(myAoIs[n][2])), 2) +
	                                               pow(float(Participants_pos[4]) - float(myAoIs[n][3]), 2)))
	                # down left
	                temp_distance.append(math.sqrt(pow(float(Participants_pos[3]) - (float(myAoIs[n][1])), 2) +
	                                               pow(float(Participants_pos[4]) - (float(myAoIs[n][3]) + float(myAoIs[n][4])), 2)))
	                # down, right
	                temp_distance.append(math.sqrt(pow(float(Participants_pos[3]) - (float(myAoIs[n][1]) + float(myAoIs[n][2])), 2) +
	                                               pow(float(Participants_pos[4]) - (float(myAoIs[n][3]) + float(myAoIs[n][4])), 2)))
	                sums_of_distances[tempAoI[m]] = sum(temp_distance)
	                break
	    # return key of minimal value in dictionary
	    return min(sums_of_distances, key=sums_of_distances.get)
	
    def getArrayRepresentationOfSequence(self, mySequences):
	    """
	    Args:
	        mySequences: String format of sequences
	    Returns: array representation of sequence
	    """
	    keys = list(mySequences.keys())
	    # odstranenie bodky na konci
	    for y in range(0, len(keys)):
	        mySequences[keys[y]] = mySequences[keys[y]].split('.')
	        del mySequences[keys[y]][len(mySequences[keys[y]]) - 1]
	    #  rozdeli D-100 na pole z dvomi prvkami D a 100
	    for y in range(0, len(keys)):
	        for z in range(0, len(mySequences[keys[y]])):
	            mySequences[keys[y]][z] = mySequences[keys[y]][z].split('-')
	    return mySequences
	
    def simplifySequence(self, aSequence):
	    """
	    Groups same fixation in a row AAABBB ->  AB and sums up the fixDur
	    Args:
	        aSequence: dictionary of array representation of sequences
	    Returns:
	        Processed sequence in array representation
	    """
	    keys = list(aSequence.keys())
	    for y in range(0, len(keys)):
	        simpleSequence = []
	        lastAOI = "0"
	        for z in range(0, len(aSequence[keys[y]])):
	            if aSequence[keys[y]][z][0] == lastAOI:
	                simpleSequence[len(simpleSequence) - 1][1] = str(int(simpleSequence[len(simpleSequence) - 1][1]) + int(aSequence[keys[y]][z][1]))
	            else:
	                simpleSequence.append([aSequence[keys[y]][z][0], aSequence[keys[y]][z][1]])
	                lastAOI = aSequence[keys[y]][z][0]
	        aSequence[keys[y]] = simpleSequence
	    return aSequence
	
	
    def applyFixDurationThreshold(self, aSequence, threshold = 80):
	    """
	    Delete fixations shorter than defined threshold
	    Args:
	        aSequence: dictionary of array representation of sequences
	        threshold: default 80 ms
	    Returns:
	        Processed sequence in array representation
	    """
	    keys = list(aSequence.keys())
	    for y in range(0, len(keys)):
	        processedArray = []
	        for z in range(0, len(aSequence[keys[y]])):
	            if int(aSequence[keys[y]][z][1]) > threshold:
	                processedArray.append(aSequence[keys[y]][z])
	        aSequence[keys[y]] = processedArray
	    return aSequence
	
    def getStringRepresentation(self, aSequence):
	    """
	    Returns string representation without duration of fixations
	    Args:
	        aSequence: dictionary of array representation of sequeces
	    """
	    newDict  = {}
	    keys = list(aSequence.keys())
	    for y in range(0, len(keys)):
	        sequence = ""
	        for z in range(0, len(aSequence[keys[y]])):
	            sequence = sequence + aSequence[keys[y]][z][0]
	        newDict[keys[y]] = sequence
	    return newDict
	
	
    def createSequencesBasedOnDistances(self, my_dataset):
	    """
	    Create Scanpath from sacade lengths(distances between fixations)
	    Args:
	        my_dataset: dataset
	    """
	    aoi_range = 300
	    sequences = {}
	    participants = my_dataset.scanpath_data_raw
	    myAoIs = my_dataset.aoi_data
	    keys = list(participants.keys())
	    for y in range(0, len(keys)):
	        sequence = ""
	        for z in range(0, min(len(participants[keys[y]]) - 1, self.dotplot_max_aois)):
	           tempdist = self.calculateDistance(int(float(participants[keys[y]][z][3])), int(float(participants[keys[y]][z][4])),
	                                        int(float(participants[keys[y]][z + 1][3])), int(float(participants[keys[y]][z + 1][4])))
	           sequence = sequence + self.getAOIBasedOnRange(tempdist, aoi_range) + "-" + str(int(float(participants[keys[y]][z + 1][1])) - int(float(participants[keys[y]][z][1]))) + "."
	        sequences[keys[y]] = sequence
	    return sequences
	
	
    def calculateDistance(self, xStart, yStart, xEnd, yEnd):
	    """
	    Calculate distance of 2D coordinates.
	    """
	    return math.sqrt(pow(xEnd - xStart, 2) + pow(yEnd - yStart, 2))
	
    def getAOIBasedOnRange(self, value, aoiRange):
	    """
	    Determine AOI based on range
	    Args:
	        value: distance between fixations
	        aoiRange: range of distance for single AOI
	    Returns: character representation of AOI
	    """
	    AOIS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
	    return AOIS[int(value / aoiRange)]
	
    def createSequencesBasedOnFixatonDurations(self, my_dataset):
	    """
	    Create Scanpath from fixations duration
	    Args:
	        my_dataset: dataset
	    """
	    aoi_range = 100
	    sequences = {}
	    participants = my_dataset.scanpath_data_raw
	    myAoIs = my_dataset.aoi_data
	    keys = list(participants.keys())
	    for y in range(0, len(keys)):
	        sequence = ""
	        for z in range(0, min(len(participants[keys[y]]) - 1, self.dotplot_max_aois)):
	           sequence = sequence + self.getAOIBasedOnRange(int(participants[keys[y]][z][2]), aoi_range) +\
	                      "-" + participants[keys[y]][z][2] + "."
	        sequences[keys[y]] = sequence
	    return sequences
	
	
    def calculateVector(self, xStart, yStart, xEnd, yEnd):
	    """
	    Calculate vector between two 2D coordinates.
	    """
	    return [xEnd - xStart, yEnd - yStart]
	
	
    def calculateAngle(self, vect1, vect2):
	    """
	    Calculates angle between 2 vector in 2D space
	    Args:
	        vect1: vector represented as list
	        vect2: vector represented as list
	    """
	    vect1Size = self.calculateDistance(0, 0, vect1[0], vect1[1])
	    vect2Size = self.calculateDistance(0, 0, vect2[0], vect2[1])
	    dotProduct = (vect1[0] * vect2[0]) + (vect1[1] * vect2[1])
	    return math.degrees(math.acos(dotProduct / (vect1Size * vect2Size)))
	
	
    def createSequencesBasedOnRelativeAngle(self, my_dataset):
	    """
	    Create Scanpath from absolute angles of saccades
	    Args:
	        my_dataset: dataset
	    """
	    aoi_range = 30
	    sequences = {}
	    participants = my_dataset.scanpath_data_raw
	    myAoIs = my_dataset.aoi_data
	    keys = list(participants.keys())
	    for y in range(0, len(keys)):
	        # TODO vec1 = vec2 at the beginning of the cycle ... better time complexity
	        # TODO add condition if sequence has two elements or so.. return empty sequence.. depends ond cycle
	        sequence = ""
	        for z in range(0, min(len(participants[keys[y]]) - 2, self.dotplot_max_aois)):
	            # calculates vector between curent point and next one
	            vec1 = self.calculateVector(int(float(participants[keys[y]][z][3])), int(float(participants[keys[y]][z][4])),
	                                   int(float(participants[keys[y]][z + 1][3])), int(float(participants[keys[y]][z + 1][4])))
	            # calculates vector between next point and next next one
	            vec2 = self.calculateVector(int(float(participants[keys[y]][z + 1][3])), int(float(participants[keys[y]][z + 1][4])),
	                                   int(float(participants[keys[y]][z + 2][3])), int(float(participants[keys[y]][z + 2][4])))
	            angle = self.calculateAngle(vec1, vec2)
	            # duration is calculated as sum of both sacades durations
	            sequence = sequence + self.getAOIBasedOnRange(angle, aoi_range) + "-" + str(int(float(participants[keys[y]][z + 2][1])) - int(float(participants[keys[y]][z][1]))) + "."
	        sequences[keys[y]] = sequence
	    return sequences
	
	
    def createSequencesBasedOnAbsoluteAngle(self, my_dataset):
	    """
	    Create Scanpath from absolute angles of saccades
	    Args:
	        my_dataset: dataset
	    """
	    aoi_range = 30
	    sequences = {}
	    participants = my_dataset.scanpath_data_raw
	    myAoIs = my_dataset.aoi_data
	    keys = list(participants.keys())
	    for y in range(0, len(keys)):
	        sequence = ""
	        for z in range(0, min(len(participants[keys[y]]) - 1, self.dotplot_max_aois)):
	            # calculates vector between curent point and next one
	            vec1 = self.calculateVector(int(float(participants[keys[y]][z][3])), int(float(participants[keys[y]][z][4])),
	                                   int(float(participants[keys[y]][z + 1][3])), int(float(participants[keys[y]][z + 1][4])))
	            # default vector (right direction)
	            vec2 = self.calculateVector(0, 0, 1, 0)
	            angle = self.calculateAngle(vec1, vec2)
	            # duration is calculated as sum of both sacades durations
	            sequence = sequence + self.getAOIBasedOnRange(angle, aoi_range) + "-" + str(int(float(participants[keys[y]][z + 1][1])) - int(float(participants[keys[y]][z][1]))) + "."
	        sequences[keys[y]] = sequence
	    return sequences
	
	
    def create_sequences_by_mod(self, my_dataset, mod):
	    """
	    Based on mod creates sequences from dataset
	    Args:
	        my_dataset: dataset
	        mod:     1 vytvori standardny scanpath z AOI
	                 2 vytvori scanpah na zaklade dlzky sakad
	                 3 vytvori scanpath na zaklade dlzky trvania fixacii
	                 4 vytvori scanpath na zaklade relativnych uhlov sakad0
	                 5 vytvori scanpath na zaklade absolutnych uhlov sakad
	    """
	
	    case = {
	      1: self.createSequencesBasedOnVisualElements,
	      2: self.createSequencesBasedOnDistances,
	      3: self.createSequencesBasedOnFixatonDurations,
	      4: self.createSequencesBasedOnRelativeAngle,
	      5: self.createSequencesBasedOnAbsoluteAngle,
	    }
	    my_func = case[mod]
	    return my_func(my_dataset)
	
	
    def dotplotListofLists(self, sequenceX, sequenceY):
	    # fill matrix with zeroes
	    dotplotMatrix = [[0 for x in sequenceX] for y in sequenceY]
	    # put 1 on matching positions
	    for xIndex, valueX in enumerate(sequenceX):
	        for yIndex, valueY in enumerate(sequenceY):
	            if valueX == valueY:
	                dotplotMatrix[yIndex][xIndex] = 1
	    return dotplotMatrix
	
	
    def findLongestCommonSequence(self, dotplotMatrix, sequenceX):
	
	    commonSubSequence = ""
	    lengthSubsequence = 0
	
	    # right part of matrix
	    for i in range(0, len(dotplotMatrix[0])):
	        # reamining x length or height of matrix
	        sum = 0
	        for j in range(0, min(len(dotplotMatrix[0]) - i, len(dotplotMatrix))):
	            sum += dotplotMatrix[j][i + j]
	
	        if sum > lengthSubsequence:
	            # sequence created by characters with value 1
	            lengthSubsequence = sum
	            commonSubSequence = ""
	            for j in range(0, min(len(dotplotMatrix[0]) - i, len(dotplotMatrix))):
	                if dotplotMatrix[j][i + j] == 1:
	                    commonSubSequence = commonSubSequence + sequenceX[i + j]
	
	
	    # left part of the matrix
	    for i in range(0, len(dotplotMatrix)):
	        sum = 0
	        for j in range(0, min(len(dotplotMatrix) - i, len(dotplotMatrix[0]))):
	            sum += dotplotMatrix[i + j][j]
	
	        if sum > lengthSubsequence:
	            # sequence created by characters with value 1
	            lengthSubsequence = sum
	            commonSubSequence = ""
	            for j in range(0, min(len(dotplotMatrix) - i, len(dotplotMatrix[0]))):
	                if dotplotMatrix[i + j][j] == 1:
	                    commonSubSequence = commonSubSequence + sequenceX[j]
	
	    return commonSubSequence
	
	
    def findCommonSequence(self, my_str_sequences):
	    """
	    Finds the most similar seqecnces in dictionary and determines their common scanpath
	    Args:
	        my_str_sequences: dictionary of sequences in string format
	    """
	    string_sequences = copy.deepcopy(my_str_sequences)
	    keys = list(string_sequences.keys())
	    while len(keys) > 1:
	        common_sequences = []
	        for y in range(0, len(keys)):
	            sequence = ""
	            for z in range(y + 1, len(keys)):
	                #  create matrix
	                matrix = self.dotplotListofLists(string_sequences[keys[y]], string_sequences[keys[z]])
	
	                # find longest common subsequence
	                subSequence = self.findLongestCommonSequence(matrix, string_sequences[keys[y]])
	                common_sequences.append([keys[y], keys[z], subSequence, len(subSequence)])
	
	        # replace 2 most similar sequences with their common sequence
	        common_sequences = sorted(common_sequences, reverse=True, key=itemgetter(3))
	        if common_sequences[0][2] == '':
	            return ''
	
	        del string_sequences[common_sequences[0][0]]
	        del string_sequences[common_sequences[0][1]]
	        string_sequences[common_sequences[0][0] + common_sequences[0][1]] = common_sequences[0][2]
	        keys = list(string_sequences.keys())
	    return string_sequences[keys[0]]
	
	
    def run_dotplot(self, simplify=True, fix_dur_threshold=None, mod=1):
	    """
	    Args:
	        raw_sequences: a Python dict of lists - {'ID1': [['F', '383'], ['G', '150']], 'ID2': .. }
	        simplify: simplify repeated fixations on the same AOI into one
	        fix_dur_threshold: minimum fixation duration threshold
	        mod: 1 - vytvori standardny scanpath z AOI
	             2 - vytvori scanpah na zaklade dlzky sakad (zatial nefunguje, nemame udaje o sakadach)
	             3 - vytvori scanpath na zaklade dlzky trvania fixacii
	             4 - vytvori scanpath na zaklade relativnych uhlov sakad (zatial nefunguje, nemame udaje o sakadach)
	             5 - vytvori scanpath na zaklade absolutnych uhlov sakad (zatial nefunguje, nemame udaje o sakadach)
	    Returns:
	        identifier: for client-side purposes
	        fixations: a list of lists representing the common scanpath - [['A', 150], ['B', 500] .. ]
	        similarity: a dict containing similarity of individual scanpaths to the common one - {'ID1': 66.66, 'ID2': ... }
	    """
	
	    # The dotplot author originally provided mods for simplifying sequences and setting their min threshold
	    # However, our system already does that automatically when uploading any scanpath data
	    """
	    my_sequences = create_sequences_by_mod(dataset_task, mod)
	    my_sequences = getArrayRepresentationOfSequence(my_sequences)
	    if fix_dur_threshold is not None:
	        my_sequences = applyFixDurationThreshold(my_sequences, fix_dur_threshold)
	    if simplify:
	        my_sequences = simplifySequence(my_sequences)
	    """
	
	    string_sequences = self.getStringRepresentation(self.raw_sequences)
	    common_scanpath_str = self.findCommonSequence(string_sequences)
	
	    common_scanpath = []
	    for fixation in common_scanpath_str:
	        common_scanpath.append([fixation, 0])
	
	    scanpath_strs = []
	    for participant_id in string_sequences:
	        scanpath_strs.append({
	            'identifier': participant_id,
	            'raw_str': string_sequences[participant_id]
	        })
	
	    res_data = {
	        'identifier': 'Dotplot',
	        'fixations': common_scanpath,
	        'similarity': similarity.calc_similarity_to_common(scanpath_strs, common_scanpath_str)
	    }
	
	    return res_data


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scikit-gaze",
    version="0.0.8",
    author="Zajic Michal",
    author_email="zajicmichal6@gmail.com",
    description="Library for scanpath comparison",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zuhi6/scikit-gaze/",
    packages=["skgaze","skgaze.core","skgaze.EMDAT","skgaze.common_scanpath","skgaze.pattern_search","skgaze.RQA","skgaze.similarity_algorithms","skgaze.RQA.CustomPythonModules"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)
